/*
 *  Copyright (c) 2020 NetApp
 *  All rights reserved
 */

package nvc

import (
	"context"
	"net/http"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	log "github.com/sirupsen/logrus"

	//Quark
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

var srv *http.Server
var (
	// Provision time for last volume by NVC
	volumeProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_provision_time",
			Help:      "Total time spent provisioning last Netapp Volume.",
		},
	)

	averageProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_avg_provision_time",
			Help:      "Average time to provision Netapp Volumes.",
		},
	)

	provisionedVolumesCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_provisioned_count_total",
			Help:      "Total count of successfully provisioned Netapp Volumes.",
		},
	)

	failedVolumesCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_failed_volume_count_total",
			Help:      "Total count of Netapp Volumes failed to provision.",
		},
	)

	knownVolumesCount = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_known_volume_count_total",
			Help:      "Total count of Netapp Volumes known to Netapp Volume Controller.",
		},
	)

	failureCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_failure_count_total",
			Help:      "Total difference in actual state and desired state of volumes for Netapp Volume Controller.",
		},
	)
)

func init() {
	prometheus.MustRegister(volumeProvisionTimeSeconds)
	prometheus.MustRegister(averageProvisionTimeSeconds)
	prometheus.MustRegister(provisionedVolumesCount)
	prometheus.MustRegister(failedVolumesCount)
	prometheus.MustRegister(knownVolumesCount)
	prometheus.MustRegister(failureCount)
}

func updateVolumeProvisionTimeSeconds(provisionTime int) {
	volumeProvisionTimeSeconds.Set(float64(provisionTime))
}

func updateAverageProvisionTimeSeconds(avgTime int) {
	averageProvisionTimeSeconds.Set(float64(avgTime))
}

func incFailedVolumesCount() {
	failedVolumesCount.Inc()
}

func incFailureCount() {
	failureCount.Inc()
}

func updateKnownVolumesCount(knownVolCount int) {
	knownVolumesCount.Set(float64(knownVolCount))
}

func incProvisionedVolumesCount() {
	provisionedVolumesCount.Inc()
}

// StartMetricsServer will start metrics exporter for nvc
func StartMetricsServer() {

	server := http.NewServeMux()
	// Handle the endpoint serving the metrics
	server.Handle("/metrics", promhttp.Handler())
	// Create a server object
	srv = &http.Server{
		Addr:    ":" + quarkconfig.NvcMetricsPort,
		Handler: server,
	}

	go func() {
		if err := srv.ListenAndServe(); err != nil && err != http.ErrServerClosed {
			log.Fatalf("Error starting nvc exporter: %s\n", err)
		}
	}()
	log.Info("NVC exporter started")
}

// StopMetricsServer will stop metrics exporter when stop signal is received
func StopMetricsServer() {
	log.Info("NVC exporter shuting down")
	ctx, _ := context.WithTimeout(context.Background(), 5*time.Second)
	if err := srv.Shutdown(ctx); err != nil {
		log.Fatalf("NVC exporter shutdown Failed: %+v", err)
	}
}
