package vbeta

import (
	"fmt"
	"strings"
	"sync"
	"time"

	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

type VolumeEvent string
type SnapshotEvent string
type CloudSnapshotEvent string
type QuarkEvent string

const (
	VolumeOnline                    VolumeEvent        = "VolOnline"
	VolumeOffline                   VolumeEvent        = "VolOffline"
	VolumeResized                   VolumeEvent        = "VolResized"
	VolumeDeleted                   VolumeEvent        = "VolDeleted"
	VolumeVersionUpgraded           VolumeEvent        = "VolVersionUpgraded"
	VolumeSnapPolicyApplied         VolumeEvent        = "VolSnapPolApplied"
	VolumeExpPolicyApplied          VolumeEvent        = "VolExpPolApplied"
	VolumeSMBEncryptionApplied      VolumeEvent        = "VolSMBEncryptionApplied"
	VolumeSMBPasswordReset          VolumeEvent        = "VolSMBPasswordReset"
	VolumeSMBDNSUpdated             VolumeEvent        = "VolSMBDNSUpdated"
	VolumeNoSnapDirOptionUpdated    VolumeEvent        = "VolumeNoSnapDirUpdated"
	SnapshotReady                   SnapshotEvent      = "SnapReady"
	SnapshotDeleted                 SnapshotEvent      = "SnapDeleted"
	SnapshotRestored                SnapshotEvent      = "SnapRestored"
	SnapshotsListUpdated            SnapshotEvent      = "SnapListUpdated"
	CloudSnapshotReady              CloudSnapshotEvent = "CloudSnapReady"
	CloudSnapshotDeleted            CloudSnapshotEvent = "CloudSnapDeleted"
	QuarkReady                      QuarkEvent         = "QuarkReady"
	QuarkNotReady                   QuarkEvent         = "QuarkNotReady"
	QuarkOperatorReady              QuarkEvent         = "QuarkOperatorReady"
	QuarkOperatorNotReady           QuarkEvent         = "QuarkOperatorNotReady"
	VolumeControllerReady           QuarkEvent         = "VolumeControllerReady"
	VolumeControllerNotReady        QuarkEvent         = "VolumeControllerNotReady"
	HighAvailabilityReady           QuarkEvent         = "HighAvailabilityReady"
	HighAvailabilityNotReady        QuarkEvent         = "HighAvailabilityNotReady"
	ExternalVolumeAccessReady       QuarkEvent         = "ExternalVolumeAccessReady"
	ExternalVolumeAccessNotReady    QuarkEvent         = "ExternalVolumeAccessNotReady"
	CloudsnapshotCapabilityReady    QuarkEvent         = "CloudsnapshotCapabilityReady"
	CloudsnapshotCapabilityNotReady QuarkEvent         = "CloudsnapshotCapabilityNotReady"
	IPDReady                        QuarkEvent         = "IPDReady"
	IPDNotReady                     QuarkEvent         = "IPDNotReady"
	EtcdClusterReady                QuarkEvent         = "EtcdClusterReady"
	EtcdClusterNotReady             QuarkEvent         = "EtcdClusterNotReady"
	NodevolReady                    QuarkEvent         = "NodevolReady"
	NodevolNotReady                 QuarkEvent         = "NodevolNotReady"
	QuarkDeleted                    QuarkEvent         = "QuarkDeleted"
	EventHandlerQueueSize                              = 100
)

type NVCClientTimeoutError struct {
	err string
}

func (te *NVCClientTimeoutError) Error() string {
	return te.err
}

func NewNVCClientTimeoutError(event string) *NVCClientTimeoutError {
	return &NVCClientTimeoutError{fmt.Sprintf("%v event timed out", event)}
}

type NVCClientFatalError struct {
	err string
}

func (fe *NVCClientFatalError) Error() string {
	return fe.err
}

func NewNVCClientFatalError(errMessage string) *NVCClientFatalError {
	return &NVCClientFatalError{errMessage}
}

type NetAppVolumeEvent struct {
	Volume *quarkv1alpha1.NetAppVolume
	//TODO: handle multiple events at the same time.
	EventType VolumeEvent
	Channel   chan error
}

type NetAppQuarkEvent struct {
	QuarkName string
	EventType QuarkEvent
	Channel   chan error
}

type NetAppVolumeSnapshotEvent struct {
	// VolumeUUID is required
	VolumeUUID string
	//SnapDisplayName is optional
	SnapDisplayName string
	//TODO: handle multiple events at the same time.
	EventType SnapshotEvent
	Channel   chan error
}

type NetAppVolumeCloudSnapshotEvent struct {
	CloudSnapshot string
	//TODO: handle multiple events at the same time.
	EventType CloudSnapshotEvent
	Channel   chan struct{}
}

type NVCClientEventHandler struct {
	// The handle to NVC Client.
	client *NVCClient
	// The channel event handler receives NetAppVolume update events from NVC Client.
	nvcVolUpdateChan <-chan *NetAppVolumeUpdate
	// The channel event handler receives NetAppVolume delete events from NVC Client.
	nvcVolDeleteChan <-chan *NetAppVolumeDelete
	// The channel event handler receives NetAppVolumeSnapshot update events from NVC Client.
	nvcSnapUpdateChan <-chan *NetAppVolumeSnapshotUpdate
	// The channel event handler receives NetAppVolumeSnapshot delete events from NVC Client.
	nvcSnapDeleteChan <-chan *NetAppVolumeSnapshotDelete
	// The channel event handler receives NetAppVolumeSnapshot restore events from NVC Client.
	nvcSnapRestoreChan <-chan *NetAppVolumeSnapshotRestore
	// The channel event handler receives NetAppVolumeSnapshotsList update events from NVC Client.
	nvcSnapsListUpdateChan <-chan *NetAppVolumeSnapshotsListUpdate
	// The channel event handler receives NetAppVolumeCloudSnapshot update events from NVC Client.
	nvcCloudSnapUpdateChan <-chan *NetAppVolumeCloudSnapshotUpdate
	// The channel event handler receives NetAppVolumeCloudSnapshot delete events from NVC Client.
	nvcCloudSnapDeleteChan <-chan *NetAppVolumeCloudSnapshotDelete
	// The channel event handler receives NetAppQuark update events from NVC Client.
	quarkUpdateChan <-chan *NetAppQuarkUpdate
	// The channel event handler receives NetAppQuark delete events from NVC Client.
	quarkDeleteChan <-chan *NetAppQuarkDelete

	// The last used subscription ID for volume events.
	volEventID int64
	// The last used subscription ID for snapshot events.
	snapEventID int64
	// The last used subscription ID for quark events.
	quarkEventID int64
	// The notification queue for NetAppVolumes.
	volSubscriptions map[int64]*NetAppVolumeEvent
	// The notification queue for NetAppVolumeSnapshots.
	snapSubscriptions map[int64]*NetAppVolumeSnapshotEvent
	// The notification queue for NetAppVolumeCloudSnapshot.
	cloudSnapSubscriptions map[int64]*NetAppVolumeCloudSnapshotEvent
	// The notification queue for Quark.
	quarkSubscriptions map[int64]*NetAppQuarkEvent
	// Event Handler mutex lock for volumes. (TODO: per-volume locks)
	volMutex sync.Mutex
	// Event Handler mutex lock for snapshots. (TODO: per-volume locks)
	snapMutex sync.Mutex
	// Event Handler mutex lock for cloud snapshots.
	cloudSnapMutex sync.Mutex
	// Event Handler mutex lock for quark.
	quarkMutex sync.Mutex
}

func newNVCClientEventHandler(client *NVCClient,
	nvcVolUpdateChan <-chan *NetAppVolumeUpdate,
	nvcVolDeleteChan <-chan *NetAppVolumeDelete,
	nvcSnapUpdateChan <-chan *NetAppVolumeSnapshotUpdate,
	nvcSnapDeleteChan <-chan *NetAppVolumeSnapshotDelete,
	nvcSnapRestoreChan <-chan *NetAppVolumeSnapshotRestore,
	nvcSnapsListUpdateChan <-chan *NetAppVolumeSnapshotsListUpdate,
	nvcCloudSnapUpdateChan <-chan *NetAppVolumeCloudSnapshotUpdate,
	nvcCloudSnapDeleteChan <-chan *NetAppVolumeCloudSnapshotDelete,
	quarkUpdateChan <-chan *NetAppQuarkUpdate,
	quarkDeleteChan <-chan *NetAppQuarkDelete) *NVCClientEventHandler {

	return &NVCClientEventHandler{
		client:                 client,
		nvcVolUpdateChan:       nvcVolUpdateChan,
		nvcVolDeleteChan:       nvcVolDeleteChan,
		nvcSnapUpdateChan:      nvcSnapUpdateChan,
		nvcSnapDeleteChan:      nvcSnapDeleteChan,
		nvcSnapRestoreChan:     nvcSnapRestoreChan,
		nvcSnapsListUpdateChan: nvcSnapsListUpdateChan,
		nvcCloudSnapUpdateChan: nvcCloudSnapUpdateChan,
		nvcCloudSnapDeleteChan: nvcCloudSnapDeleteChan,
		quarkUpdateChan:        quarkUpdateChan,
		quarkDeleteChan:        quarkDeleteChan,
		volEventID:             1,
		snapEventID:            1,
		volSubscriptions:       make(map[int64]*NetAppVolumeEvent),
		snapSubscriptions:      make(map[int64]*NetAppVolumeSnapshotEvent),
		cloudSnapSubscriptions: make(map[int64]*NetAppVolumeCloudSnapshotEvent),
		quarkSubscriptions:     make(map[int64]*NetAppQuarkEvent),
		volMutex:               sync.Mutex{},
		snapMutex:              sync.Mutex{},
		cloudSnapMutex:         sync.Mutex{},
		quarkMutex:             sync.Mutex{},
	}
}

func (h *NVCClientEventHandler) Run(wg *sync.WaitGroup) {
	defer wg.Done()
	// Instead of reuse, we create an new WaitGroup to track internal
	// goroutines.
	handlerWg := sync.WaitGroup{}
	defer handlerWg.Wait()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case volUpdate, ok := <-h.nvcVolUpdateChan:
				if !ok {
					log.Info("NVC Client volume update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"volume":                 volUpdate.Volume.Name,
					"resized":                volUpdate.Status.SizeUpdated,
					"exportPolicyApplied":    volUpdate.Status.ExportPolicyUpdated,
					"snapshotPolicyApplied":  volUpdate.Status.SnapshotPolicyUpdated,
					"smbEncryptionApplied":   volUpdate.Status.SMBEncryptionUpdated,
					"smbDNSUpdated":          volUpdate.Status.SMBDNSUpdated,
					"smbPasswordReset":       volUpdate.Status.SMBPasswordReset,
					"NoSnapDirOptionUpdated": volUpdate.Status.NoSnapDirOptionUpdated,
					"conditions":             volUpdate.UpdatedConditions,
				}).Debug("Received a NetAppVolume update event.")
				h.volMutex.Lock()
				for id, sub := range h.volSubscriptions {
					if sub.Volume.Name == volUpdate.Volume.Name {
						// First, we pass fatal errors through the event channel.
						// Second, we notify of the subscribed events by closing
						// the channel.
						switch sub.EventType {
						case VolumeOnline:
							if !IsVolumeOnline(volUpdate.UpdatedConditions) {
								// Check if there's a fatal error corresponding to this event.
								// this error could be a retriable error which is considered fatal only after timeout
								errMsg := h.volumeOnlineError(volUpdate.UpdatedConditions)
								if errMsg != "" {
									// only send fatal error on channel
									if !strings.Contains(strings.ToLower(errMsg), "retrying") {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeOffline:
							if !IsVolumeOnline(volUpdate.UpdatedConditions) {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeResized:
							if !volUpdate.Status.SizeUpdated {
								// this error could be a retriable error which is considered fatal only after timeout
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeResizeError)
								if errMsg != "" {
									if !strings.Contains(strings.ToLower(errMsg), "retrying") {
										// only send fatal error on channel
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeSnapPolicyApplied:
							if !volUpdate.Status.SnapshotPolicyUpdated {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeExportPolicyError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeExpPolicyApplied:
							if !volUpdate.Status.ExportPolicyUpdated {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeExportPolicyError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeSMBEncryptionApplied:
							if !volUpdate.Status.SMBEncryptionUpdated {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeSMBConfigurationError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeSMBDNSUpdated:
							if volUpdate.Status.SMBDNSUpdated {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeNoSnapDirOptionUpdated:
							if volUpdate.Status.NoSnapDirOptionUpdated {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeSMBPasswordReset:
							if !volUpdate.Status.SMBPasswordReset {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeSMBPasswordResetError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeVersionUpgraded:
							if !volUpdate.Status.QuarkVersionUpdated {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeUpgradeFailed)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						}
					}
				}
				h.volMutex.Unlock()
			case vol, ok := <-h.nvcVolDeleteChan:
				if !ok {
					log.Info("NVC Client volume delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"volume": vol.Volume.Name,
				}).Debug("Received a NetAppVolume delete event.")
				h.volMutex.Lock()
				for id, sub := range h.volSubscriptions {
					if sub.Volume.Name == vol.Volume.Name {
						switch sub.EventType {
						case VolumeDeleted:
							close(sub.Channel)
							delete(h.volSubscriptions, id)
						}
					}
				}
				h.volMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case snapUpdate, ok := <-h.nvcSnapUpdateChan:
				if !ok {
					log.Info("NVC Client snapshot update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"snapshot":   snapUpdate.Snapshot.Name,
					"conditions": snapUpdate.UpdatedConditions,
				}).Debug("Received a NetAppVolumeSnapshot update event.")
				h.snapMutex.Lock()
				for id, sub := range h.snapSubscriptions {
					if sub.SnapDisplayName == snapUpdate.Snapshot.Spec.DisplayName &&
						snapUpdate.Snapshot.ObjectMeta.GetLabels()[quarkconfig.VolumeLabel] == sub.VolumeUUID {
						switch sub.EventType {
						case SnapshotReady:
							if !IsSnapshotReady(snapUpdate.UpdatedConditions) {
								errMsg := h.snapshotError(snapUpdate.UpdatedConditions, quarkv1alpha1.SnapshotCreateError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.snapSubscriptions, id)
							}
						}
					}
				}
				h.snapMutex.Unlock()
			case snapDelete, ok := <-h.nvcSnapDeleteChan:
				if !ok {
					log.Info("NVC Client snapshot delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"deletedSnapshots": snapDelete.DeletedSnapshots,
				}).Debug("Received a NetAppVolumeSnapshot delete event.")
				h.snapMutex.Lock()
				for id, sub := range h.snapSubscriptions {
					if snapDelete.VolumeUUID == sub.VolumeUUID &&
						util.StringSetContains(snapDelete.DeletedSnapshots, sub.SnapDisplayName) {
						switch sub.EventType {
						case SnapshotDeleted:
							close(sub.Channel)
							delete(h.snapSubscriptions, id)
						}
					}
				}
				h.snapMutex.Unlock()
			case snapRestore, ok := <-h.nvcSnapRestoreChan:
				if !ok {
					log.Info("NVC Client snapshot restore event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"volume":           snapRestore.VolumeUUID,
					"errorCondition":   snapRestore.ErrorCondition,
					"restoredSnapshot": snapRestore.RestoredSnapshot,
				}).Debug("Received a NetAppVolumeSnapshot restore event.")
				h.snapMutex.Lock()
				for id, sub := range h.snapSubscriptions {
					if snapRestore.VolumeUUID == sub.VolumeUUID {
						switch sub.EventType {
						case SnapshotRestored:
							if !snapRestore.RestoredSnapshot {
								errMsg := snapRestore.ErrorCondition
								if errMsg != "" {
									if !strings.Contains(strings.ToLower(errMsg), "retrying") {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.snapSubscriptions, id)
							}
						}
					}
				}
				h.snapMutex.Unlock()
			case snapListUpdate, ok := <-h.nvcSnapsListUpdateChan:
				if !ok {
					log.Info("NVC Client snapshotsList update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"volume":               snapListUpdate.VolumeUUID,
					"snapshotsListUpdated": snapListUpdate.SnapshotsListUpdated,
				}).Debug("Received a NetAppVolumeSnapshotsListUpdate event.")
				h.snapMutex.Lock()
				for id, sub := range h.snapSubscriptions {
					if snapListUpdate.VolumeUUID == sub.VolumeUUID &&
						snapListUpdate.SnapshotsListUpdated {
						switch sub.EventType {
						case SnapshotsListUpdated:
							close(sub.Channel)
							delete(h.snapSubscriptions, id)
						}
					}
				}
				h.snapMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case cloudSnapUpdate, ok := <-h.nvcCloudSnapUpdateChan:
				if !ok {
					log.Info("NVC Client cloud snapshot update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"cloudSnapshot":     cloudSnapUpdate.Status.CloudSnapshotUUID,
					"ready":             cloudSnapUpdate.Status.Ready,
					"uploading":         cloudSnapUpdate.Status.Uploading,
					"deleting":          cloudSnapUpdate.Status.Deleting,
					"endpointUUID":      cloudSnapUpdate.Status.EndpointUUID,
					"bytesTransferred":  cloudSnapUpdate.Status.BytesTransferred,
					"logicalSize":       cloudSnapUpdate.Status.LogicalSize,
					"completionPercent": cloudSnapUpdate.Status.CompletionPercent,
					"error":             cloudSnapUpdate.Status.Error,
					"creationTime":      cloudSnapUpdate.Status.CreationTime,
				}).Debug("Received a NetAppVolumeCloudSnapshot update event.")
				h.volMutex.Lock()
				for id, sub := range h.cloudSnapSubscriptions {
					if sub.CloudSnapshot == cloudSnapUpdate.CloudSnapshot.Name {
						switch sub.EventType {
						case CloudSnapshotReady:
							if cloudSnapUpdate.Status.Ready {
								close(sub.Channel)
								delete(h.cloudSnapSubscriptions, id)
							}
						}
					}
				}
				h.volMutex.Unlock()
			case cloudSnapDelete, ok := <-h.nvcCloudSnapDeleteChan:
				if !ok {
					log.Info("NVC Client cloud snapshot delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"cloudSnapshot": cloudSnapDelete.CloudSnapshot.Name,
				}).Debug("Received a NetAppVolumeCloudSnapshot delete event.")
				h.volMutex.Lock()
				for id, sub := range h.cloudSnapSubscriptions {
					if sub.CloudSnapshot == cloudSnapDelete.CloudSnapshot.Name {
						switch sub.EventType {
						case CloudSnapshotDeleted:
							close(sub.Channel)
							delete(h.cloudSnapSubscriptions, id)
						}
					}
				}
				h.volMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case quarkUpdate, ok := <-h.quarkUpdateChan:
				if !ok {
					log.Info("NVC Client quark update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"quark":                 quarkUpdate.Quark.Name,
					"volumeControllerReady": quarkUpdate.Status.VolumeControllerReady,
					"highAvailabilityReady": quarkUpdate.Status.HighAvailabilityReady,
					"operatorReady":         quarkUpdate.Status.OperatorReady,
					"version":               quarkUpdate.Status.Version,
				}).Debug("Received a Quark update event.")
				h.quarkMutex.Lock()
				for id, sub := range h.quarkSubscriptions {
					if sub.QuarkName == quarkUpdate.Quark.Name {
						switch sub.EventType {
						case QuarkReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if quarkUpdate.Status.VolumeControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case QuarkNotReady:
							if !quarkUpdate.Status.VolumeControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case QuarkOperatorReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if quarkUpdate.Status.OperatorReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case QuarkOperatorNotReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if !quarkUpdate.Status.OperatorReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case VolumeControllerReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if quarkUpdate.Status.VolumeControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case VolumeControllerNotReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if !quarkUpdate.Status.VolumeControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case HighAvailabilityReady:
							if quarkUpdate.Status.HighAvailabilityReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case HighAvailabilityNotReady:
							if !quarkUpdate.Status.HighAvailabilityReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case ExternalVolumeAccessReady:
							if quarkUpdate.Status.ExternalVolumeAccessReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case ExternalVolumeAccessNotReady:
							if !quarkUpdate.Status.ExternalVolumeAccessReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case CloudsnapshotCapabilityReady:
							if quarkUpdate.Status.CloudsnapshotCapabilityReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case CloudsnapshotCapabilityNotReady:
							if !quarkUpdate.Status.CloudsnapshotCapabilityReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case IPDReady:
							if quarkUpdate.Status.IPDReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case IPDNotReady:
							if !quarkUpdate.Status.IPDReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case EtcdClusterReady:
							if quarkUpdate.Status.EtcdClusterReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case EtcdClusterNotReady:
							if !quarkUpdate.Status.EtcdClusterReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case NodevolReady:
							if quarkUpdate.Status.NodevolReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case NodevolNotReady:
							if !quarkUpdate.Status.NodevolReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						}
					}
				}
				h.quarkMutex.Unlock()
			case quarkDelete, ok := <-h.quarkDeleteChan:
				if !ok {
					log.Info("NVC Client quark delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"quark": quarkDelete.Quark.Name,
				}).Debug("Received a Quark delete event.")
				h.quarkMutex.Lock()
				for id, sub := range h.quarkSubscriptions {
					if sub.QuarkName == quarkDelete.Quark.Name {
						switch sub.EventType {
						case QuarkDeleted:
							// TODO: Add NVCClientFatalError if needed to channel
							close(sub.Channel)
							delete(h.quarkSubscriptions, id)
						}
					}
				}
				h.quarkMutex.Unlock()
			}
		}
	}()
}

func (h *NVCClientEventHandler) SubscribeVolumeEvent(event *NetAppVolumeEvent) int64 {
	h.volMutex.Lock()
	defer h.volMutex.Unlock()
	vol, err := h.client.NetAppVolumeGet(event.Volume.Name)
	if vol != nil && err == nil {
		switch event.EventType {
		case VolumeOnline:
			if h.client.NetAppVolumeOnline(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeOffline:
			if !h.client.NetAppVolumeOnline(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeResized:
			if h.client.NetAppVolumeResized(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeExpPolicyApplied:
			if h.client.NetAppVolExportPolicyUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeSnapPolicyApplied:
			if h.client.NetAppVolSnapPolicyUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeNoSnapDirOptionUpdated:
			if h.client.NetAppVolNoSnapDirOptionUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeSMBEncryptionApplied:
			if h.client.NetAppVolSMBEncryptionUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeSMBDNSUpdated:
			if h.client.NetAppVolSMBDNSUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeSMBPasswordReset:
			if h.client.NetAppVolSMBPasswordResetDone(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeVersionUpgraded:
			if h.client.NetAppVolumeUpgraded(vol) {
				close(event.Channel)
				return 0
			}
		}
	} else {
		switch event.EventType {
		case VolumeDeleted:
			close(event.Channel)
			return 0
		}
	}
	h.volSubscriptions[h.volEventID] = event
	ret := h.volEventID
	h.volEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeVolumeEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.volMutex.Lock()
	close(h.volSubscriptions[eventID].Channel)
	delete(h.volSubscriptions, eventID)
	h.volMutex.Unlock()
}

// WaitVolumeEventWithTimeout is a blocking call that returns when a volume
// event of interest is seen, when the event handler times out, or when the
// operation that relates to the event of interest has encountered a fatal
// failure.
func (h *NVCClientEventHandler) WaitVolumeEventWithTimeout(
	volume *quarkv1alpha1.NetAppVolume,
	eventType VolumeEvent,
	timeout time.Duration) error {
	volEvent := &NetAppVolumeEvent{
		Volume:    volume,
		EventType: eventType,
		Channel:   make(chan error),
	}
	// return fatal error if occurred before subscribing
	retryErr, err := h.processVolumeEventError(volume.Name, eventType)
	if err != nil && !retryErr {
		return err
	}
	subID := h.SubscribeVolumeEvent(volEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-volEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeVolumeEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				h.UnsubscribeVolumeEvent(subID)
				_, err := h.processVolumeEventError(volume.Name, eventType)
				if err != nil {
					return err
				}
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

func (h *NVCClientEventHandler) SubscribeQuarkEvent(event *NetAppQuarkEvent) int64 {
	h.quarkMutex.Lock()
	defer h.quarkMutex.Unlock()
	quark, err := h.client.NetAppQuarkGet(event.QuarkName)
	if quark != nil && err == nil {
		switch event.EventType {
		case QuarkReady:
			if h.client.NetAppQuarkReady(quark) {
				close(event.Channel)
				return 0
			}
		case QuarkNotReady:
			if !h.client.NetAppQuarkReady(quark) {
				close(event.Channel)
				return 0
			}
		case QuarkOperatorReady:
			if h.client.NetAppQuarkOperatorReady(quark) {
				close(event.Channel)
				return 0
			}
		case QuarkOperatorNotReady:
			if !h.client.NetAppQuarkOperatorReady(quark) {
				close(event.Channel)
				return 0
			}
		case VolumeControllerReady:
			if status := getNetAppQuarkStatus(quark); status.VolumeControllerReady {
				close(event.Channel)
				return 0
			}
		case VolumeControllerNotReady:
			if status := getNetAppQuarkStatus(quark); !status.VolumeControllerReady {
				close(event.Channel)
				return 0
			}
		case HighAvailabilityReady:
			if status := getNetAppQuarkStatus(quark); status.HighAvailabilityReady {
				close(event.Channel)
				return 0
			}
		case HighAvailabilityNotReady:
			if status := getNetAppQuarkStatus(quark); !status.HighAvailabilityReady {
				close(event.Channel)
				return 0
			}
		case ExternalVolumeAccessReady:
			if status := getNetAppQuarkStatus(quark); status.ExternalVolumeAccessReady {
				close(event.Channel)
				return 0
			}
		case ExternalVolumeAccessNotReady:
			if status := getNetAppQuarkStatus(quark); !status.ExternalVolumeAccessReady {
				close(event.Channel)
				return 0
			}
		case CloudsnapshotCapabilityReady:
			if status := getNetAppQuarkStatus(quark); status.CloudsnapshotCapabilityReady {
				close(event.Channel)
				return 0
			}
		case CloudsnapshotCapabilityNotReady:
			if status := getNetAppQuarkStatus(quark); !status.CloudsnapshotCapabilityReady {
				close(event.Channel)
				return 0
			}
		case IPDReady:
			if status := getNetAppQuarkStatus(quark); status.IPDReady {
				close(event.Channel)
				return 0
			}
		case IPDNotReady:
			if status := getNetAppQuarkStatus(quark); !status.IPDReady {
				close(event.Channel)
				return 0
			}
		case EtcdClusterReady:
			if status := getNetAppQuarkStatus(quark); status.EtcdClusterReady {
				close(event.Channel)
				return 0
			}
		case EtcdClusterNotReady:
			if status := getNetAppQuarkStatus(quark); !status.EtcdClusterReady {
				close(event.Channel)
				return 0
			}
		case NodevolReady:
			if status := getNetAppQuarkStatus(quark); status.NodevolReady {
				close(event.Channel)
				return 0
			}
		case NodevolNotReady:
			if status := getNetAppQuarkStatus(quark); !status.NodevolReady {
				close(event.Channel)
				return 0
			}
		}
	} else {
		switch event.EventType {
		case QuarkDeleted:
			close(event.Channel)
			return 0
		}
	}
	h.quarkSubscriptions[h.quarkEventID] = event
	ret := h.quarkEventID
	h.quarkEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeQuarkEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.quarkMutex.Lock()
	close(h.quarkSubscriptions[eventID].Channel)
	delete(h.quarkSubscriptions, eventID)
	h.quarkMutex.Unlock()
}

// WaitQuarkEventWithTimeout is a blocking call that returns when a quark
// event of interest is seen, when the event handler times out, or when the
// operation that relates to the event of interest has encountered a fatal
// failure.
func (h *NVCClientEventHandler) WaitQuarkEventWithTimeout(
	quarkName string,
	eventType QuarkEvent,
	timeout time.Duration) error {
	quarkEvent := &NetAppQuarkEvent{
		QuarkName: quarkName,
		EventType: eventType,
		Channel:   make(chan error),
	}

	subID := h.SubscribeQuarkEvent(quarkEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-quarkEvent.Channel:
				if ok {
					// TODO: Return NVCClientFatalError if raised
					return err
				}
				return nil
			case <-time.After(timeout):
				h.UnsubscribeQuarkEvent(subID)
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

func (h *NVCClientEventHandler) SubscribeSnapshotEvent(event *NetAppVolumeSnapshotEvent) int64 {
	h.snapMutex.Lock()
	defer h.snapMutex.Unlock()
	switch event.EventType {
	case SnapshotReady:
		if h.client.NetAppVolumeSnapshotReady(event.VolumeUUID, event.SnapDisplayName) {
			close(event.Channel)
			return 0
		}
	case SnapshotDeleted:
		if h.client.NetAppVolumeSnapshotDeleted(event.VolumeUUID, event.SnapDisplayName) {
			close(event.Channel)
			return 0
		}
	case SnapshotRestored:
		if h.client.NetAppVolumeSnapshotRestored(event.VolumeUUID, event.SnapDisplayName) {
			close(event.Channel)
			return 0
		}
	}

	h.snapSubscriptions[h.snapEventID] = event
	ret := h.snapEventID
	h.snapEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeSnapshotEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.snapMutex.Lock()
	close(h.snapSubscriptions[eventID].Channel)
	delete(h.snapSubscriptions, eventID)
	h.snapMutex.Unlock()
}

// WaitSnapshotEventWithTimeout is a blocking call that returns when a snapshot
// event of interest is seen, when the event handler times out, or when the
// operation that relates to the event of interest has encountered a fatal
// failure.
func (h *NVCClientEventHandler) WaitSnapshotEventWithTimeout(
	volName string, snapDisplayName string,
	eventType SnapshotEvent,
	timeout time.Duration) error {
	snapEvent := &NetAppVolumeSnapshotEvent{
		VolumeUUID:      volName,
		SnapDisplayName: snapDisplayName,
		EventType:       eventType,
		Channel:         make(chan error),
	}
	// return fatal error if occurred before subscribing
	retryErr, err := h.processSnapshotEventError(volName, eventType)
	if err != nil && !retryErr {
		return err
	}
	subID := h.SubscribeSnapshotEvent(snapEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-snapEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeSnapshotEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				h.UnsubscribeSnapshotEvent(subID)
				_, err := h.processSnapshotEventError(volName, eventType)
				if err != nil {
					return err
				}
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

func (h *NVCClientEventHandler) SubscribeCloudSnapshotEvent(event *NetAppVolumeCloudSnapshotEvent) int64 {
	h.cloudSnapMutex.Lock()
	defer h.cloudSnapMutex.Unlock()
	switch event.EventType {
	case CloudSnapshotReady:
		if h.client.NetAppVolumeCloudSnapshotReady(event.CloudSnapshot) {
			close(event.Channel)
			return 0
		}
	case CloudSnapshotDeleted:
		if h.client.NetAppVolumeCloudSnapshotDeleted(event.CloudSnapshot) {
			close(event.Channel)
			return 0
		}
	}

	h.cloudSnapSubscriptions[h.snapEventID] = event
	ret := h.snapEventID
	h.snapEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeCloudSnapshotEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.cloudSnapMutex.Lock()
	close(h.cloudSnapSubscriptions[eventID].Channel)
	delete(h.cloudSnapSubscriptions, eventID)
	h.cloudSnapMutex.Unlock()
}

// IsVolumeOnline returns true if a volume is online.
func IsVolumeOnline(conditions []quarkv1alpha1.NetAppVolumeCondition) bool {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppVolumeOnline && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// IsSnapshotReady returns true if a snapshot is ready.
func IsSnapshotReady(conditions []quarkv1alpha1.NetAppVolumeSnapshotCondition) bool {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.SnapshotReady && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// processes error if present in NetAppVolume conditions
func (h *NVCClientEventHandler) processVolumeEventError(volumeUUID string, eventType VolumeEvent) (retryErr bool, ret error) {
	var errMsg string
	conditions, _ := h.client.NetAppVolumeGetConditions(volumeUUID)
	if conditions != nil {
		switch eventType {
		case VolumeOnline:
			errMsg = h.volumeOnlineError(conditions)
		case VolumeResized:
			errMsg = h.volumeUpdateError(conditions, quarkv1alpha1.NetAppVolumeResizeError)
		case VolumeSMBEncryptionApplied:
			errMsg = h.volumeUpdateError(conditions, quarkv1alpha1.NetAppVolumeSMBConfigurationError)
		case VolumeExpPolicyApplied:
			errMsg = h.volumeUpdateError(conditions, quarkv1alpha1.NetAppVolumeExportPolicyError)
		case VolumeSnapPolicyApplied:
			errMsg = h.volumeUpdateError(conditions, quarkv1alpha1.NetAppVolumeSnapshotPolicyError)
		}
	}
	if errMsg != "" {
		ret, retryErr = createFatalError(errMsg)
	}
	return retryErr, ret
}

// processes error if present in NetAppVolumeSnapshot or NetAppVolumeSnapshotsList conditions
func (h *NVCClientEventHandler) processSnapshotEventError(volumeUUID string, eventType SnapshotEvent) (retryErr bool, ret error) {
	var errMsg string
	switch eventType {
	case SnapshotRestored:
		conditions, _ := h.client.NetAppVolumeSnapshotsListGetConditions(volumeUUID)
		if conditions != nil {
			errMsg = h.snapRestoreError(conditions)
		}
	}
	if errMsg != "" {
		ret, retryErr = createFatalError(errMsg)
	}
	return retryErr, ret
}

// returns volume online error message
func (h *NVCClientEventHandler) volumeOnlineError(conditions []quarkv1alpha1.NetAppVolumeCondition) string {
	for _, cond := range conditions {
		if util.StringSetContains(getFatalVolumeOnlineErrors(), string(cond.Type)) && cond.Status == corev1.ConditionTrue {
			return cond.Message
		}
	}
	return ""
}

func (h *NVCClientEventHandler) volumeUpdateError(conditions []quarkv1alpha1.NetAppVolumeCondition, errType quarkv1alpha1.NetAppVolumeConditionType) string {
	for _, cond := range conditions {
		if cond.Type == errType && cond.Status == corev1.ConditionTrue {
			return cond.Message
		}
	}
	return ""
}

// returns snap restore error message
func (h *NVCClientEventHandler) snapRestoreError(conditions []quarkv1alpha1.NetAppVolumeSnapshotsListCondition) string {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.SnapshotRestoreError && cond.Status == corev1.ConditionTrue {
			return cond.Message
		}
	}
	return ""
}

// returns snapshot related error message
func (h *NVCClientEventHandler) snapshotError(conditions []quarkv1alpha1.NetAppVolumeSnapshotCondition, errType quarkv1alpha1.NetAppVolumeSnapshotConditionType) string {
	for _, cond := range conditions {
		if cond.Type == errType && cond.Status == corev1.ConditionTrue {
			return cond.Message
		}
	}
	return ""
}

// constructs fatal error from error message
func createFatalError(errMsg string) (err *NVCClientFatalError, retryErr bool) {
	if strings.Contains(strings.ToLower(errMsg), "retrying") {
		msgSlice := strings.Split(errMsg, ":")
		// to get payload message, obtain message after ':' as retry errors will be of the format "<HTTP status> Retrying: <payload message>"
		// TODO: add retry message creation in util package to be used by podrick and other packages for ensuring consistency in error messages
		err = NewNVCClientFatalError(strings.TrimSpace(msgSlice[len(msgSlice)-1]))
		retryErr = true
		return
	}
	err = NewNVCClientFatalError(errMsg)
	return
}

func getFatalVolumeOnlineErrors() []string {
	return []string{string(quarkv1alpha1.NetAppVolumeSMBConfigurationError), string(quarkv1alpha1.NetAppVolumeExportPolicyError), string(quarkv1alpha1.NetAppVolumeSnapshotPolicyError)}
}
