package vbeta

import (
	"context"
	"encoding/json"
	"fmt"
	"net"
	"sort"
	"strings"
	"time"

	uuid "github.com/satori/go.uuid"
	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/resource"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime/schema"
	utilruntime "k8s.io/apimachinery/pkg/util/runtime"

	apiextensionsclient "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
	"k8s.io/client-go/kubernetes"
	_ "k8s.io/client-go/plugin/pkg/client/auth"
	"k8s.io/client-go/rest"
	"k8s.io/client-go/tools/cache"
	"k8s.io/client-go/tools/clientcmd"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"

	quarkinformers "quark.netapp.io/volume-controller/pkg/client/informers/externalversions"
	quarklisters "quark.netapp.io/volume-controller/pkg/client/listers/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/nvc"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

var (
	// enforcing 1 Gi volume increment for beta version and onwards
	incrementalSize resource.Quantity = resource.MustParse("1Gi")
	minVolSize      resource.Quantity = resource.MustParse("100Gi")
	maxVolSize      resource.Quantity = resource.MustParse("100Ti")
)

const (
	versionUpgradeErr = "Failed to upgrade NetAppVolume"
	upgradeAbortErr   = "Failed to abort NetAppVolumeUpgrade"
)

type StaleVersionError struct {
}

func (e StaleVersionError) Error() string {
	return "StaleObjectVersion"
}

type NVCClient struct {
	// Kubernetes clientset
	kubeClient kubernetes.Interface
	// apiextensionsclient is the clientset for Crds
	crdClient apiextensionsclient.Interface
	// Quark clientset
	quarkClient quarkclientset.Interface
	// Kubernetes cluster config
	kubeConfig *rest.Config
	// NVC namespace (i.e., where NVC custom resources will be created)
	nvcNamespace string
	// Client event channels
	volumeDeleteChannel        chan<- *NetAppVolumeDelete
	volumeUpdateChannel        chan<- *NetAppVolumeUpdate
	snapshotDeleteChannel      chan<- *NetAppVolumeSnapshotDelete
	snapshotUpdateChannel      chan<- *NetAppVolumeSnapshotUpdate
	snapshotRestoreChannel     chan<- *NetAppVolumeSnapshotRestore
	snapshotsListChannel       chan<- *NetAppVolumeSnapshotsListUpdate
	cloudSnapshotUpdateChannel chan<- *NetAppVolumeCloudSnapshotUpdate
	cloudSnapshotDeleteChannel chan<- *NetAppVolumeCloudSnapshotDelete
	quarkUpdateChannel         chan<- *NetAppQuarkUpdate
	quarkDeleteChannel         chan<- *NetAppQuarkDelete

	// Controller parameters to monitor NetAppVolumes
	volumesLister        quarklisters.NetAppVolumeLister
	volumesSynced        cache.InformerSynced
	snapshotsLister      quarklisters.NetAppVolumeSnapshotLister
	snapshotsSynced      cache.InformerSynced
	snapshotsListLister  quarklisters.NetAppVolumeSnapshotsListLister
	snapshotsListSynced  cache.InformerSynced
	cloudSnapshotsLister quarklisters.NetAppVolumeCloudSnapshotLister
	cloudSnapshotsSynced cache.InformerSynced
	quarksLister         quarklisters.QuarkLister
	quarksSynced         cache.InformerSynced
	//workqueue              workqueue.RateLimitingInterface
	//workqueueSnapshot      workqueue.RateLimitingInterface
	//workqueueSnapshotsList workqueue.RateLimitingInterface
	//knownVolumes             map[string]bool
	//knownVolumeSnapshots     map[string]bool
	//knownVolumeSnapshotsList map[string]bool
	DarkContent bool
}

type NetAppVolumeUsage quarkv1alpha1.NetAppVolumeUsageStatus

func newNVCClient(kubeConfig *rest.Config, namespace string) (*NVCClient, error) {
	// Setting up clientsets
	kubeClient, err := kubernetes.NewForConfig(kubeConfig)
	if err != nil {
		return nil, err
	}
	crdClient, err := apiextensionsclient.NewForConfig(kubeConfig)
	if err != nil {
		return nil, err
	}
	quarkClient, err := quarkclientset.NewForConfig(kubeConfig)
	if err != nil {
		return nil, err
	}

	client := &NVCClient{
		kubeClient:   kubeClient,
		crdClient:    crdClient,
		quarkClient:  quarkClient,
		kubeConfig:   kubeConfig,
		nvcNamespace: namespace,
		//knownVolumes:             make(map[string]bool),
		//knownVolumeSnapshots:     make(map[string]bool),
		//knownVolumeSnapshotsList: make(map[string]bool),
	}
	return client, nil
}

// Creates and initializes NVCClient
//kubeConfigPath: Path to cluster config file
//apiServerIP: Api server IP, overrides kubeConfigPath if specified
//namespace: Namespace where NVC is deployed
func NewNVCClient(kubeConfigPath, apiServerIP, namespace string) (*NVCClient, error) {
	kubeConfig, err := clientcmd.BuildConfigFromFlags(apiServerIP, kubeConfigPath)
	if err != nil {
		return nil, err
	}
	return newNVCClient(kubeConfig, namespace)
}

// Creates and initializes NVCClient
// kubeConfigPath: Path to cluster config file
// apiServerIP: API server IP, overrides kubeConfigPath if specified
// namespace: Namespace where NVC is deployed
// minVolSize: Minimum supported volume size
func NewNVCClientWithSize(kubeConfigPath, apiServerIP, namespace string, minVolSizeString string) (client *NVCClient, err error) {
	minVolSize, err = resource.ParseQuantity(minVolSizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}

	return NewNVCClient(kubeConfigPath, apiServerIP, namespace)
}

// Creates and initializes NVCClient
// kubeConfigBytes: kubeconfig content
// namespace: Namespace where NVC is deployed
func NewNVCClientWithKubeConfigBytes(kubeConfigBytes []byte, namespace string) (*NVCClient, error) {
	kubeConfig, err := clientcmd.RESTConfigFromKubeConfig(kubeConfigBytes)
	if err != nil {
		return nil, err
	}
	return newNVCClient(kubeConfig, namespace)
}

// Creates and initializes NVCClient
// kubeConfigPath: Path to cluster config file
// apiServerIP: Api server IP, overrides kubeConfigPath if specified
// namespace: Namespace where NVC is deployed
// darkContentFlag: flag to enable dark content
func NewNVCClientDark(kubeConfigPath, apiServerIP, namespace string, darkContentFlag bool) (*NVCClient, error) {
	client, err := NewNVCClient(kubeConfigPath, apiServerIP, namespace)
	if err == nil {
		client.SetDarkContent(darkContentFlag)
	}
	return client, err
}

// Creates and initializes NVCClient
// kubeConfigPath: Path to cluster config file
// apiServerIP: API server IP, overrides kubeConfigPath if specified
// namespace: Namespace where NVC is deployed
// minVolSize: Minimum supported volume size
// darkContentFlag: flag to enable dark content
func NewNVCClientWithSizeDark(kubeConfigPath, apiServerIP, namespace string, minVolSizeString string, darkContentFlag bool) (*NVCClient, error) {
	client, err := NewNVCClientWithSize(kubeConfigPath, apiServerIP, namespace, minVolSizeString)
	if err == nil {
		client.SetDarkContent(darkContentFlag)
	}
	return client, err
}

// Creates and initializes NVCClient
// kubeConfigBytes: kubeconfig content
// namespace: Namespace where NVC is deployed
// darkContentFlag: flag to enable dark content
func NewNVCClientWithKubeConfigBytesDark(kubeConfigBytes []byte, namespace string, darkContentFlag bool) (*NVCClient, error) {

	client, err := NewNVCClientWithKubeConfigBytes(kubeConfigBytes, namespace)
	if err == nil {
		client.SetDarkContent(darkContentFlag)
	}
	return client, err
}

// SetDarkContent sets the value for the "dark content" flag
func (client *NVCClient) SetDarkContent(val bool) {
	client.DarkContent = val
}

func (client *NVCClient) NetAppVolumeCreate(volume *quarkv1alpha1.NetAppVolume) (*quarkv1alpha1.NetAppVolume, error) {

	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("failed to get Quark: %s", err.Error())
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("no quarks found")
	}
	quark := quarks[0]
	// add quark version to create NetAppVolume only for "Production" & "TestVariant" configs
	switch strings.ToLower(quark.Spec.ControllerConfig.Variant) {
	case quarkconfig.ProductionVariant, quarkconfig.OnPremVariant:
		volume.Spec.QuarkVersion = quark.Status.Version
	}

	return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Create(
		context.TODO(), volume, metav1.CreateOptions{})
}

func (client *NVCClient) validateADCredentials(user, password string) error {
	if user == "" {
		return fmt.Errorf("username cannot be empty")
	}
	if len(user) > 104 {
		return fmt.Errorf("username cannot be more than 104 characters long")
	}
	if password == "" {
		return fmt.Errorf("password cannot be empty")
	}
	return nil
}

// Creates AD Secret without ownerRef set
func (client *NVCClient) createOrUpdateADSecret(secretName, user, password string) (*corev1.Secret, error) {
	err := client.validateADCredentials(user, password)
	if err != nil {
		return nil, err
	}

	secretData := map[string][]byte{
		"username": []byte(user),
		"password": []byte(password),
	}
	secret := &corev1.Secret{
		TypeMeta: metav1.TypeMeta{},
		Data:     secretData,
		Type:     corev1.SecretTypeOpaque,
	}
	secret.Name = secretName
	secret.Namespace = client.nvcNamespace
	return client.SecretCreateOrUpdate(secret)
}

// Sets ownerRef field of secret
func (client *NVCClient) setSecretOwnerRef(secretName string, volume *quarkv1alpha1.NetAppVolume) (*corev1.Secret, error) {
	secret, err := client.SecretGet(secretName)
	if err != nil {
		return nil, err
	}
	secret.ObjectMeta = nvc.MakeObjectMeta(volume)
	secret.Name = secretName

	return client.SecretUpdate(secret)

}

// Creates AD Secret with ownerRef set
func (client *NVCClient) createOrUpdateADSecretWithOwnerRef(volume *quarkv1alpha1.NetAppVolume, secretName, user, password string) (*corev1.Secret, error) {
	err := client.validateADCredentials(user, password)
	if err != nil {
		return nil, err
	}

	secretData := map[string][]byte{
		"username": []byte(user),
		"password": []byte(password),
	}
	secret := &corev1.Secret{
		TypeMeta:   metav1.TypeMeta{},
		ObjectMeta: nvc.MakeObjectMeta(volume),
		Data:       secretData,
		Type:       corev1.SecretTypeOpaque,
	}
	secret.Name = secretName
	secret.Namespace = client.nvcNamespace

	return client.SecretCreateOrUpdate(secret)
}

// Initiate NetAppVolume version upgrade
func (client *NVCClient) NetAppVolumeUpgrade(volumeUUID, version string) (*quarkv1alpha1.NetAppVolume, error) {
	volume, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil && errors.IsNotFound(err) {
		return nil, fmt.Errorf("NetappVolume %s was not found for upgrade", volumeUUID)
	}
	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("%s: %s", versionUpgradeErr, err)
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("%s: no quarks found", versionUpgradeErr)
	}
	quark := quarks[0]
	if quark.Status.Version != version {
		return nil, fmt.Errorf("%s: quark is not updated to version %s", versionUpgradeErr, version)
	}
	// initiate upgrade when Quark is Ready and updated with required version for volume upgrade
	volume.Spec.QuarkVersion = version
	newVol, err := client.netAppVolumeUpdate(volume)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return newVol, nil
}

// Abort NetAppVolume version upgrade
func (client *NVCClient) NetAppVolumeUpgradeAbort(volumeUUID string) (*quarkv1alpha1.NetAppVolume, error) {
	volume, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil && errors.IsNotFound(err) {
		return nil, fmt.Errorf("NetappVolume %s was not found for aborting upgrade", volumeUUID)
	}
	if volume.Spec.QuarkVersion == "" {
		return nil, fmt.Errorf("%s: NetAppVolume Spec version is missing", upgradeAbortErr)
	}
	// check if abort is possible
	if volume.Status.QuarkVersion != "" && volume.Spec.QuarkVersion != volume.Status.QuarkVersion {
		err := client.checkUpgradeAbort(volume.Status.Conditions)
		if err != nil {
			return nil, fmt.Errorf("%s: %s", upgradeAbortErr, err.Error())
		}
	}
	// revert to old version, in order to signal abort
	volume.Spec.QuarkVersion = volume.Status.QuarkVersion
	newVolume, err := client.netAppVolumeUpdate(volume)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, fmt.Errorf("%s: %s", upgradeAbortErr, err.Error())
	}
	return newVolume, nil
}

// returns error if abort cannot proceed
func (client *NVCClient) checkUpgradeAbort(conditions []quarkv1alpha1.NetAppVolumeCondition) error {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppVolumeUpgradeDMAPHaltSucceeded {
			// do not proceed with abort if halt succeeded
			return fmt.Errorf("deployment is being updated for volume upgrade")
		}
	}
	return nil
}

// Creates SMB Volume
func (client *NVCClient) NetAppVolumeCreateSMB(vol *quarkv1alpha1.NetAppVolume, user, password string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	secretName := util.GetVolCreateADSecretName(vol.Name)
	_, err = client.createOrUpdateADSecret(secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": vol.Name,
			"secret": secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	volume, err = client.NetAppVolumeCreate(vol)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": vol.Name,
			"secret": secretName,
		}).Error("Failed to create volume.")
		return nil, err
	}

	_, err = client.setSecretOwnerRef(secretName, volume)
	// If secret is already deleted ignore the error.
	if err != nil && !errors.IsNotFound(err) {
		// Delete Secret and Volume if can't set ownerRef
		_ = client.SecretDelete(secretName)
		_ = client.NetAppVolumeDelete(volume.Name)
		log.WithFields(log.Fields{
			"volume": vol.Name,
			"secret": secretName,
		}).Error("Failed to set OwnerRef field.")

		return nil, err
	}
	return volume, nil
}

func (client *NVCClient) NetAppVolumeDeleteSMB(volumeUUID string, user, password string) error {
	volume, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		if errors.IsNotFound(err) {
			return nil
		} else {
			return err
		}
	}
	_, err = client.createOrUpdateADSecretWithOwnerRef(volume, util.GetVolDeleteADSecretName(volume.Name), user, password)
	if err != nil {
		return err
	}
	return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Delete(
		context.TODO(), volumeUUID, metav1.DeleteOptions{})
}

func (client *NVCClient) NetAppVolumeGet(volumeUUID string) (*quarkv1alpha1.NetAppVolume, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
}

func (client *NVCClient) NetAppVolumeGetVersion(volumeUUID string) (string, error) {
	volume, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return "", fmt.Errorf("failed to get the NetappVolume version: %s", err.Error())
	}
	return volume.Spec.QuarkVersion, nil
}

// returns conditions from NetAppVolume CR status
func (client *NVCClient) NetAppVolumeGetConditions(volumeUUID string) ([]quarkv1alpha1.NetAppVolumeCondition, error) {
	volume, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetappVolume: %s", err.Error())
	}
	return volume.Status.Conditions, nil
}

// returns conditions from NetAppVolumeSnapshotsList CR status
func (client *NVCClient) NetAppVolumeSnapshotsListGetConditions(volumeUUID string) ([]quarkv1alpha1.NetAppVolumeSnapshotsListCondition, error) {
	snapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetappVolumeSnapshotsList: %s", err.Error())
	}
	return snapList.Status.Conditions, nil
}

func (client *NVCClient) NetAppVolumeGetAll() ([]*quarkv1alpha1.NetAppVolume, error) {
	volumes := make([]*quarkv1alpha1.NetAppVolume, 0)
	volumesList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	for i := range volumesList.Items {
		volumes = append(volumes, &volumesList.Items[i])
	}
	return volumes, err
}

func (client *NVCClient) NetAppVolumeGetAllSMB() ([]*quarkv1alpha1.NetAppVolume, error) {
	volumes := make([]*quarkv1alpha1.NetAppVolume, 0)
	volumesList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	for i := range volumesList.Items {
		if volumesList.Items[i].Spec.Protocol == "cifs" {
			volumes = append(volumes, &volumesList.Items[i])
		}
	}
	return volumes, err
}

func (client *NVCClient) NetAppVolumeDelete(volumeUUID string) error {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Delete(
		context.TODO(), volumeUUID, metav1.DeleteOptions{})
}

func (client *NVCClient) NetAppVolumeDeleteAll() error {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).DeleteCollection(
		context.TODO(), metav1.DeleteOptions{}, metav1.ListOptions{})
}

func (client *NVCClient) netAppVolumeUpdate(volume *quarkv1alpha1.NetAppVolume) (*quarkv1alpha1.NetAppVolume, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Update(
		context.TODO(), volume, metav1.UpdateOptions{})
}

// NetAppVolumeOnline returns "true" when the Quark pod is ready to serve data.
func (client *NVCClient) NetAppVolumeOnline(volume *quarkv1alpha1.NetAppVolume) bool {
	for _, cond := range volume.Status.Conditions {
		if cond.Type == quarkv1alpha1.NetAppVolumeOnline && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// NetAppVolumeResizeAbort aborts the current Vol-Resize operation
func (client *NVCClient) NetAppVolumeResizeAbort(volumeUUID string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	// Reset the new size in spec. The volume error conditions can then be cleared by Podrick.
	volume.Spec.Size = volume.Status.Size
	volume, err = client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Update(
		context.TODO(), volume, metav1.UpdateOptions{})
	if err != nil {
		return nil, err
	}
	return volume, nil
}

// NetAppVolumeUsage returns the usage statistics of a given volume.
func (client *NVCClient) NetAppVolumeUsage(volume *quarkv1alpha1.NetAppVolume) (*NetAppVolumeUsage, error) {
	usage := &NetAppVolumeUsage{}
	volUsage, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeUsages(client.nvcNamespace).Get(
		context.TODO(), volume.Name, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	usage.LogicalUsed = volUsage.Status.LogicalUsed
	usage.PhysicalUsed = volUsage.Status.PhysicalUsed
	usage.AllocatedSize = volUsage.Status.AllocatedSize
	usage.SnapshotUsed = volUsage.Status.SnapshotUsed
	return usage, nil
}

// NetAppVolumeResized returns "true" when the data PVCs are correctly resized.
func (client *NVCClient) NetAppVolumeResized(volume *quarkv1alpha1.NetAppVolume) bool {
	return volume.Status.Size.Cmp(volume.Spec.Size) == 0
}

// NetAppVolExportPolicyUpdated returns "true" when the ExportPolicy is applied to the volume.
func (client *NVCClient) NetAppVolExportPolicyUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.ExportPolicy == volume.Spec.ExportPolicy)
}

// NetAppVolSnapPolicyUpdated returns "true" when the SnapshotPolicy is applied to the volume.
func (client *NVCClient) NetAppVolSnapPolicyUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.SnapshotPolicy == volume.Spec.SnapshotPolicy)
}

// NetAppVolNoSnapOptionUpdated returns "true" when the NoSnapDir is applied to the volume.
func (client *NVCClient) NetAppVolNoSnapDirOptionUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.NoSnapDir == volume.Spec.NoSnapDir)
}

// NetAppVolSMBEncryptionUpdated returns "true" when SMB Encryption is applied to the volume.
func (client *NVCClient) NetAppVolSMBEncryptionUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Spec.SMBConfiguration.EncryptionEnabled == volume.Status.SMBConfiguration.EncryptionEnabled)
}

// NetAppVolSMBDNSUpdated returns "true" when SMB DNSIPs are applied to the volume.
func (client *NVCClient) NetAppVolSMBDNSUpdated(volume *quarkv1alpha1.NetAppVolume) bool {
	return util.IsEqualStrings(volume.Spec.SMBConfiguration.DNSIPs, volume.Status.SMBConfiguration.DNSIPs)
}

// NetAppVolSMBPasswordResetDone returns "true" when machine-account password reset is successful.
func (client *NVCClient) NetAppVolSMBPasswordResetDone(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.SMBPasswordResetCount == volume.Spec.SMBPasswordResetCount)
}

// NetAppVolumeUpgraded returns "true" when Volume version upgrade is successful.
func (client *NVCClient) NetAppVolumeUpgraded(volume *quarkv1alpha1.NetAppVolume) bool {
	conditions := volume.Status.Conditions
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppVolumeUpgradeDMAPHaltSucceeded || cond.Type == quarkv1alpha1.NetAppVolumeUpgradeDMAPHaltRequested || cond.Type == quarkv1alpha1.NetAppVolumeUpgradeFailed {
			// return false if another upgrade is in progress
			return false
		}
	}
	return (volume.Status.QuarkVersion == volume.Spec.QuarkVersion)

}

// NetAppVolumeCloudSnapshotRestoreCondition returns the state of a cloud snapshot restore operation.
func (client *NVCClient) NetAppVolumeCloudSnapshotRestoreCondition(volume *quarkv1alpha1.NetAppVolume) *quarkv1alpha1.NetAppVolumeCondition {
	for _, condition := range volume.Status.Conditions {
		if condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotPreparing ||
			condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestoring ||
			condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestored ||
			condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestoreError {

			return condition.DeepCopy()
		}
	}
	return nil
}

func (client *NVCClient) NetAppExportPolicyCreate(policy *quarkv1alpha1.NetAppExportPolicy) (*quarkv1alpha1.NetAppExportPolicy, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppExportPolicies(client.nvcNamespace).Create(
		context.TODO(), policy, metav1.CreateOptions{})
}

func (client *NVCClient) NetAppExportPolicyGet(policyName string) (*quarkv1alpha1.NetAppExportPolicy, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppExportPolicies(client.nvcNamespace).Get(
		context.TODO(), policyName, metav1.GetOptions{})
}

func (client *NVCClient) NetAppExportPolicyDelete(policyName string) error {
	return client.quarkClient.QuarkV1alpha1().NetAppExportPolicies(client.nvcNamespace).Delete(
		context.TODO(), policyName, metav1.DeleteOptions{})
}

func (client *NVCClient) NetAppSnapshotPolicyCreate(policy *quarkv1alpha1.NetAppSnapshotPolicy) (*quarkv1alpha1.NetAppSnapshotPolicy, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppSnapshotPolicies(client.nvcNamespace).Create(
		context.TODO(), policy, metav1.CreateOptions{})
}

func (client *NVCClient) NetAppSnapshotPolicyGet(policyName string) (*quarkv1alpha1.NetAppSnapshotPolicy, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppSnapshotPolicies(client.nvcNamespace).Get(
		context.TODO(), policyName, metav1.GetOptions{})
}

func (client *NVCClient) NetAppSnapshotPolicyDelete(policyName string) error {
	return client.quarkClient.QuarkV1alpha1().NetAppSnapshotPolicies(client.nvcNamespace).Delete(
		context.TODO(), policyName, metav1.DeleteOptions{})
}

func (client *NVCClient) NetAppVolumeSnapshotCreate(volumeSnapshot *quarkv1alpha1.NetAppVolumeSnapshot) error {
	metaLabels := volumeSnapshot.ObjectMeta.GetLabels()
	volumeUUID := metaLabels[quarkconfig.VolumeLabel]
	volume, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return err
	}

	// if snapshot with same display name already exists in snapshotList CR, do not allow to create snapshot
	snapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s", volumeUUID)
	}

	for _, snap := range snapList.Status.Snapshots {
		if volumeSnapshot.Spec.DisplayName == snap.Name {
			return fmt.Errorf("failed to create snapshot with same name as existing snapshot %s", snap.Name)
		}
	}

	volumeSnapshot.ObjectMeta.OwnerReferences = []metav1.OwnerReference{
		*metav1.NewControllerRef(volume, schema.GroupVersionKind{
			Group:   quarkv1alpha1.SchemeGroupVersion.Group,
			Version: quarkv1alpha1.SchemeGroupVersion.Version,
			Kind:    "NetAppVolume",
		}),
	}

	_, err = client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshots(client.nvcNamespace).Create(
		context.TODO(), volumeSnapshot, metav1.CreateOptions{})
	if err != nil {
		return err
	}
	return nil
}

// NetAppVolumeSnapshotCreateAbort aborts the current snapshot create operation
func (client *NVCClient) NetAppVolumeSnapshotCreateAbort(volumeUUID, snapUUID string) error {

	// Check if the snapshot is ready
	snapReady, err := client.netAppVolumeSnapshotStatusReady(snapUUID, volumeUUID)
	if snapReady {
		return fmt.Errorf("unable to abort snapshot create %s  for volume %s , snapshot already created", snapUUID, volumeUUID)
	} else if err != nil {
		return fmt.Errorf("unable to check snapshot status for snapshot %s : %v", snapUUID, err)
	}

	// snapshot not ready we can proceed with abort , delete snapshot CR
	err = client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshots(client.nvcNamespace).Delete(
		context.TODO(), snapUUID, metav1.DeleteOptions{})
	if err != nil {
		return err
	}

	log.WithFields(log.Fields{
		"volume":   volumeUUID,
		"snapshot": snapUUID,
	}).Debug("NetAppVolumeSnapshots create abort success.")

	return nil
}

func (client *NVCClient) NetAppVolumeSnapshotDelete(volumeUUID, snapDisplayName string) error {
	snapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s when deleting snapshot %s", volumeUUID, snapDisplayName)
	}
	snapListCopy := snapList.DeepCopy()
	for _, snap := range snapListCopy.Status.Snapshots {
		if snap.Name == snapDisplayName {
			for {
				pendingDeleteSnapshots := util.StringSetDifference(snapListCopy.Spec.DeletedSnapshots, snapListCopy.Status.DeletedSnapshots)
				if util.StringSetContains(pendingDeleteSnapshots, snapDisplayName) {
					log.WithFields(log.Fields{
						"volume":       volumeUUID,
						"snapshot":     snapDisplayName,
						"snapshotList": snapListCopy,
					}).Debug("Snapshot delete operation is already pending. Returning success.")
					return nil
				}
				// Update spec.DeletedSnapshots field of snapshotsList CR to delete a snapshot
				// and clear already deleted snapshots from both spec and status.
				snapListCopy.Spec.DeletedSnapshots = append(pendingDeleteSnapshots, snapDisplayName)
				// Trim from status the snapshots that are not part of spec
				// anymore (just in case spec and status get out of sync).
				snapListCopy.Status.DeletedSnapshots = make([]string, 0)

				_, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Update(
					context.TODO(), snapListCopy, metav1.UpdateOptions{})

				if err != nil {
					if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
						snapListCopy, err = client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
							context.TODO(), volumeUUID, metav1.GetOptions{})
						if err != nil {
							return fmt.Errorf("server error deleting snapshot %s:%s", snapDisplayName, err)
						}
						continue
					}
					return fmt.Errorf("server error deleting snapshot %s: %s", snapDisplayName, err)
				}
				break
			}
			return nil
		}
	}
	// Return nil, if snapshot not found, to make delete operation idempotent
	return nil
}

// NetAppVolumeSnapshotDeleteAbort aborts the current snapshot delete operation
// This call is not idempotent , if this is called for an already aborted deletion , it will error out
func (client *NVCClient) NetAppVolumeSnapshotDeleteAbort(volumeUUID, snapDisplayName string) error {
	errPrefix := "could not abort snapshot delete operation"
	snapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s when aborting Snapshot Delete", volumeUUID)
	}
	snapListCopy := snapList.DeepCopy()

	for _, snap := range snapListCopy.Status.Snapshots {
		if snap.Name == snapDisplayName {
			for true {

				// Check if the snapshot has already been deleted
				if util.StringSetContains(snapListCopy.Status.DeletedSnapshots, snapDisplayName) {
					// Snapshot has already been deleted, cannot abort deletion, return error
					return fmt.Errorf(errPrefix+", snapshot %s has already been deleted", snapDisplayName)
				}

				if !util.StringSetContains(snapListCopy.Spec.DeletedSnapshots, snapDisplayName) {
					// This means a previous abort might have succeeded or a snapshot Deletion never started
					// It is difficult to determine the state , so return error
					return fmt.Errorf(errPrefix+", there is no pending delete for snapshot %s", snapDisplayName)
				}

				// Remove snapshot from DeletedSnapshots list in Spec
				snapListCopy.Spec.DeletedSnapshots = util.RemoveStringFromSlice(snapListCopy.Spec.DeletedSnapshots, snapDisplayName)

				_, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Update(
					context.TODO(), snapListCopy, metav1.UpdateOptions{})
				if err != nil {
					if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
						snapListCopy, err = client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
							context.TODO(), volumeUUID, metav1.GetOptions{})
						if err != nil {
							return fmt.Errorf(errPrefix+" for %s server error:%s", snapDisplayName, err)
						}
						continue
					}
					return fmt.Errorf(errPrefix+" for %s server error:%s", snapDisplayName, err)
				}
				break
			}
			return nil
		}
	}

	// If the snapshotname is not found return an error
	return fmt.Errorf("The snapshot deletion to be aborted %s has already been deleted or doesn't exist", snapDisplayName)
}

func (client *NVCClient) NetAppVolumeSnapshotRestore(volumeUUID, snapDisplayName string) error {
	snapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s when restoring snapshot %s", volumeUUID, snapDisplayName)
	}
	snapListCopy := snapList.DeepCopy()
	for _, snap := range snapListCopy.Status.Snapshots {
		// Check if the snapshot to be restored exists
		if snap.Name == snapDisplayName {
			for {
				// if spec and status do not match and error condition exists in snapshotsList conditions,
				// send the message to abort restore operation
				// if there is no SnapshotRestoreError condition present, indicate that the restore is pending.
				if snapListCopy.Spec.RestoredSnapshot != snapListCopy.Status.RestoredSnapshot {
					for _, cond := range snapList.Status.Conditions {
						if cond.Type == quarkv1alpha1.SnapshotRestoreError && cond.Status == corev1.ConditionTrue {
							log.WithFields(log.Fields{
								"volume":       volumeUUID,
								"snapshot":     snapDisplayName,
								"snapshotList": snapListCopy,
							}).Error("Fatal error occurred during SnapRestore.")
							return fmt.Errorf("fatal error occurred during SnapRestore of %s. Manually abort this operation to trigger another SnapRestore", snapDisplayName)
						}
					}
					if snapListCopy.Spec.RestoredSnapshot != snapDisplayName {
						log.WithFields(log.Fields{
							"volume":       volumeUUID,
							"snapshot":     snapDisplayName,
							"snapshotList": snapListCopy,
						}).Debug("Another snapshot restore operation is already pending.")
						return fmt.Errorf("another snapshot restore operation is already pending")
					}
					log.WithFields(log.Fields{
						"volume":       volumeUUID,
						"snapshot":     snapDisplayName,
						"snapshotList": snapListCopy,
					}).Debug("Snapshot restore operation is pending, returning success")
					return nil
				}
				// clear the status of already restored snapshot and update the spec with new
				// snapshot to be restored
				snapListCopy.Spec.RestoredSnapshot = snapDisplayName
				snapListCopy.Status.RestoredSnapshot = ""
				_, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Update(
					context.TODO(), snapListCopy, metav1.UpdateOptions{})
				if err != nil {
					if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
						snapListCopy, err = client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
							context.TODO(), volumeUUID, metav1.GetOptions{})
						if err != nil {
							return fmt.Errorf("failed to get snapshots list :%s", err)
						}
						continue
					}
					return fmt.Errorf("server error restoring snapshot %s: %s", snapDisplayName, err)
				}
				break
			}
			return nil
		}
	}
	// Return error, if snapshot not found for restore
	return fmt.Errorf("snapshot not found for restore: %s", snapDisplayName)
}

// NetAppVolumeSnapshotRestoreAbort aborts the current snapRestore operation
func (client *NVCClient) NetAppVolumeSnapshotRestoreAbort(volumeUUID string) error {
	snapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return fmt.Errorf("failed to retrieve snapshot list info for volume %s when aborting Snap Restore", volumeUUID)
	}
	// Updating conditions by removing the SnapshotRestoreError condition
	var newConditions []quarkv1alpha1.NetAppVolumeSnapshotsListCondition
	for _, cond := range snapList.Status.Conditions {
		if cond.Type != quarkv1alpha1.SnapshotRestoreError {
			newConditions = append(newConditions, cond)
		}
	}
	snapListCopy := snapList.DeepCopy()
	snapListCopy.Spec.RestoredSnapshot = ""
	snapListCopy.Status.RestoredSnapshot = ""
	snapListCopy.Status.Conditions = newConditions
	_, err = client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Update(
		context.TODO(), snapListCopy, metav1.UpdateOptions{})
	if err != nil {
		return fmt.Errorf("server error aborting NetAppVolumeSnapshotRestore: %s", err)
	}
	log.WithFields(log.Fields{
		"volume": volumeUUID,
	}).Info("NetAppVolumeSnapshotRestore has been aborted.")

	return nil
}

// NetAppVolumeSnapshotRestored returns "true" when the snapshot is restored.
func (client *NVCClient) NetAppVolumeSnapshotRestored(volumeUUID, snapDisplayName string) bool {
	snapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return false
	}
	// If spec and status match and snapshot is present in the RestoredSnapshot of snapshoList CR's status, it has been restored, hence return true
	if snapList.Spec.RestoredSnapshot == snapList.Status.RestoredSnapshot && snapList.Status.RestoredSnapshot == snapDisplayName {
		log.WithFields(log.Fields{
			"volume":   volumeUUID,
			"snapshot": snapDisplayName,
		}).Debug("NetAppVolumeSnapshot has been restored.")
		return true
	}
	log.WithFields(log.Fields{
		"volume":   volumeUUID,
		"snapshot": snapDisplayName,
	}).Debug("NetAppVolumeSnapshot is not restored.")
	return false
}

// NetAppVolumeSnapshotDeleted returns "true" when the snapshot is deleted.
func (client *NVCClient) NetAppVolumeSnapshotDeleted(volumeUUID, snapDisplayName string) bool {
	_, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		// if volume is not found, then there is no snapshot
		return true
	}
	snapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return false
	}
	// If snapshot is not found in the list, it is deleted
	for _, snap := range snapList.Status.Snapshots {
		if snap.Name == snapDisplayName {
			log.WithFields(log.Fields{
				"volume":   volumeUUID,
				"snapshot": snapDisplayName,
			}).Debug("NetAppVolumeSnapshot isn't deleted yet.")
			return false
		}
	}
	log.WithFields(log.Fields{
		"volume":   volumeUUID,
		"snapshot": snapDisplayName,
	}).Debug("NetAppVolumeSnapshot is deleted.")
	return true
}

type NetAppVolumeSnapshotInfo struct {
	Name       string
	Size       string
	VolumeUUID string
	// Creation time of snapshot with location set to UTC.
	CreationTime time.Time
}

func (client *NVCClient) NetAppVolumeSnapshotsListGetByVolume(volumeUUID string) ([]NetAppVolumeSnapshotInfo, error) {
	var snapList []NetAppVolumeSnapshotInfo
	_, err := client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		if errors.IsNotFound(err) {
			return nil, fmt.Errorf("netappVolume %s was not found", volumeUUID)
		}
	}
	ntapVolSnapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("server error listing snapshots")
	}

	for _, snap := range ntapVolSnapList.Status.Snapshots {
		snapList = append(snapList, NetAppVolumeSnapshotInfo{
			Name:         snap.Name,
			Size:         snap.Size.String(),
			VolumeUUID:   volumeUUID,
			CreationTime: snap.CreationTime.UTC()})
	}
	return snapList, nil
}

func (client *NVCClient) NetAppVolumeSnapshotGetAll() ([]NetAppVolumeSnapshotInfo, error) {
	var snapList []NetAppVolumeSnapshotInfo
	ntapVols, err := client.NetAppVolumeGetAll()
	if err != nil {
		return nil, err
	}
	for _, volume := range ntapVols {
		ntapVolSnapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
			context.TODO(), volume.Name, metav1.GetOptions{})
		if err != nil {
			return nil, fmt.Errorf("server error listing snapshots")
		}

		for _, snap := range ntapVolSnapList.Status.Snapshots {
			snapList = append(snapList, NetAppVolumeSnapshotInfo{
				Name:       snap.Name,
				Size:       snap.Size.String(),
				VolumeUUID: volume.Name})
		}

	}
	return snapList, nil
}

// NetAppVolumeSnapshotReady returns "true" when the volume snapshot is ready.
func (client *NVCClient) NetAppVolumeSnapshotReady(volumeUUID, snapDisplayName string) bool {
	ntapVolSnapList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshotsLists(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return false
	}
	log.WithFields(log.Fields{
		"volume":    volumeUUID,
		"snapshots": ntapVolSnapList,
	}).Debug("NetAppVolumeSnapshotsList get success.")

	// If snapshot found in the list, it is ready
	for _, snap := range ntapVolSnapList.Status.Snapshots {
		if snap.Name == snapDisplayName {
			return true
		}
	}
	return false
}

// NetAppVolumeSnapshotStatusReady returns "true" when the volume snapshot CR status is ready.
func (client *NVCClient) netAppVolumeSnapshotStatusReady(snapUUID, volumeUUID string) (bool, error) {
	ntapVolSnapshot, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeSnapshots(client.nvcNamespace).Get(
		context.TODO(), snapUUID, metav1.GetOptions{})
	if err != nil {
		return false, err
	}
	log.WithFields(log.Fields{
		"volume":    volumeUUID,
		"snapshots": ntapVolSnapshot,
	}).Debug("NetAppVolumeSnapshots get success.")

	if len(ntapVolSnapshot.Status.Conditions) != 0 && IsSnapshotReady(ntapVolSnapshot.Status.Conditions) {
		//Snapshot is ready
		return true, nil
	}

	return false, nil
}

// NetAppVolumeDelete is used to signal the deletion of a NetAppVolume.
type NetAppVolumeDelete struct {
	Volume *quarkv1alpha1.NetAppVolume
}

// NetAppVolumeUpdate is used to signal the most recent state of a NetAppVolume and the
// most recent changes.
type NetAppVolumeUpdate struct {
	Volume            *quarkv1alpha1.NetAppVolume
	Status            *NetAppVolumeUpdateStatus
	UpdatedConditions []quarkv1alpha1.NetAppVolumeCondition
	Error             error
}

// NetAppQuarkUpdate is used to signal the most recent state of a Quark CR and the
// most recent changes.
type NetAppQuarkUpdate struct {
	Quark  *quarkv1alpha1.Quark
	Status NetAppQuarkUpdateStatus
	Error  error
}

// NetAppQuarkDelete is used to signal the most recent state of a Quark CR and the
// most recent changes.
type NetAppQuarkDelete struct {
	Quark *quarkv1alpha1.Quark
}

// QuarkOperatorUpdateStatus captures the status of any pending updates.
type NetAppQuarkUpdateStatus struct {
	HighAvailabilityReady        bool
	ExternalVolumeAccessReady    bool
	CloudsnapshotCapabilityReady bool
	VolumeControllerReady        bool
	OperatorReady                bool
	IPDReady                     bool
	EtcdClusterReady             bool
	NodevolReady                 bool
	Version                      string
}

// NetAppVolumeUpdateStatus captures the status of any pending updates.
type NetAppVolumeUpdateStatus struct {
	SizeUpdated            bool
	ExportPolicyUpdated    bool
	SnapshotPolicyUpdated  bool
	SMBEncryptionUpdated   bool
	SMBDNSUpdated          bool
	SMBPasswordReset       bool
	QuarkVersionUpdated    bool
	NoSnapDirOptionUpdated bool
}

// NetAppVolumeSnapshotDelete is used to signal the snapshot delete completion.
type NetAppVolumeSnapshotDelete struct {
	VolumeUUID       string
	DeletedSnapshots []string
}

// NetAppVolumeSnapshotUpdate is used to signal the most recent changes to NetAppVolumeSnapshots.
type NetAppVolumeSnapshotUpdate struct {
	Snapshot          *quarkv1alpha1.NetAppVolumeSnapshot
	UpdatedConditions []quarkv1alpha1.NetAppVolumeSnapshotCondition
}

// NetAppVolumeSnapshotRestore is used to signal the snapshot restore operation.
type NetAppVolumeSnapshotRestore struct {
	VolumeUUID       string
	ErrorCondition   string
	RestoredSnapshot bool
}

// NetAppVolumeSnapshotsListUpdate is used to signal the most recent changes to NetAppVolumeSnapshotsList.
type NetAppVolumeSnapshotsListUpdate struct {
	VolumeUUID           string
	SnapshotsListUpdated bool
}

// NetAppVolumeCloudSnapshotDelete is used to signal the deletion of a NetAppVolumeCloudSnapshot.
type NetAppVolumeCloudSnapshotDelete struct {
	CloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot
}

// NetAppVolumeCloudSnapshotUpdate is used to signal the most recent state of a NetAppVolumeCloudSnapshot and the
// most recent changes.
type NetAppVolumeCloudSnapshotUpdate struct {
	CloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot
	Status        NetAppVolumeCloudSnapshotInfo
}

// NetAppVolumeUpdated returns "true" when the actual state has converged to the
// desired state. If not, it will return an object describing what actions are pending.
func (client *NVCClient) NetAppVolumeUpdated(volume *quarkv1alpha1.NetAppVolume) (bool, *NetAppVolumeUpdateStatus, error) {
	statusConverged := true

	status := &NetAppVolumeUpdateStatus{NoSnapDirOptionUpdated: true, SizeUpdated: true, ExportPolicyUpdated: true, SnapshotPolicyUpdated: true, SMBEncryptionUpdated: true, SMBDNSUpdated: true, SMBPasswordReset: true, QuarkVersionUpdated: true}
	// Check size
	if !client.NetAppVolumeResized(volume) {
		status.SizeUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolExportPolicyUpdated(volume) {
		status.ExportPolicyUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolSnapPolicyUpdated(volume) {
		status.SnapshotPolicyUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolSMBEncryptionUpdated(volume) {
		status.SMBEncryptionUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolSMBDNSUpdated(volume) {
		status.SMBDNSUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolSMBPasswordResetDone(volume) {
		status.SMBPasswordReset = false
		statusConverged = false
	}
	if !client.NetAppVolumeUpgraded(volume) {
		status.QuarkVersionUpdated = false
		statusConverged = false
	}
	if !client.NetAppVolNoSnapDirOptionUpdated(volume) {
		status.NoSnapDirOptionUpdated = false
		statusConverged = false
	}
	// Return the status of any potentially pending updates.
	if statusConverged {
		return true, status, nil
	} else {
		return false, status, nil
	}
}

// Start the NVC Client controller to receive updates from NVC.
// stopCh: Channel to terminate the controller.
// volUpdateChan: Channel to receive updates to NetAppVolumes.
// volDeleteChan: Channel to receive deletions of NetAppVolumes.
// snapUpdateChan: Channel to receive updates to NetAppVolumeSnapshots.
// snapDeleteChan: Channel to receive deletions of NetAppVolumeSnapshots.
// quarkUpdateChan: Channel to receive updates to Quark
// quarkDeleteChan: Channel to receive deletions of Quark
// threadiness: Number of controller threads
func (client *NVCClient) StartController(stopCh <-chan struct{}, threadiness int) (*NVCClientEventHandler, error) {

	volUpdateChan := make(chan *NetAppVolumeUpdate, EventHandlerQueueSize)
	volDeleteChan := make(chan *NetAppVolumeDelete, EventHandlerQueueSize)
	snapDeleteChan := make(chan *NetAppVolumeSnapshotDelete, EventHandlerQueueSize)
	snapRestoreChan := make(chan *NetAppVolumeSnapshotRestore, EventHandlerQueueSize)
	snapUpdateChan := make(chan *NetAppVolumeSnapshotUpdate, EventHandlerQueueSize)
	snapListChan := make(chan *NetAppVolumeSnapshotsListUpdate, EventHandlerQueueSize)
	cloudSnapUpdateChan := make(chan *NetAppVolumeCloudSnapshotUpdate, EventHandlerQueueSize)
	cloudSnapDeleteChan := make(chan *NetAppVolumeCloudSnapshotDelete, EventHandlerQueueSize)
	quarkUpdateChan := make(chan *NetAppQuarkUpdate, EventHandlerQueueSize)
	quarkDeleteChan := make(chan *NetAppQuarkDelete, EventHandlerQueueSize)

	eventHandler := newNVCClientEventHandler(client, volUpdateChan, volDeleteChan,
		snapUpdateChan, snapDeleteChan, snapRestoreChan, snapListChan,
		cloudSnapUpdateChan, cloudSnapDeleteChan, quarkUpdateChan, quarkDeleteChan)

	client.volumeUpdateChannel = volUpdateChan
	client.volumeDeleteChannel = volDeleteChan
	client.snapshotUpdateChannel = snapUpdateChan
	client.snapshotDeleteChannel = snapDeleteChan
	client.snapshotRestoreChannel = snapRestoreChan
	client.snapshotsListChannel = snapListChan
	client.cloudSnapshotUpdateChannel = cloudSnapUpdateChan
	client.cloudSnapshotDeleteChannel = cloudSnapDeleteChan
	client.quarkUpdateChannel = quarkUpdateChan
	client.quarkDeleteChannel = quarkDeleteChan

	if err := client.setupController(stopCh); err != nil {
		return nil, err
	}

	//go client.runNetAppVolumeController(threadiness, stopCh)
	return eventHandler, nil
}

// Stops the event controller.
func (client *NVCClient) StopController() {
	close(client.volumeUpdateChannel)
	client.volumeUpdateChannel = nil
	close(client.snapshotUpdateChannel)
	client.snapshotUpdateChannel = nil
	close(client.volumeDeleteChannel)
	client.volumeDeleteChannel = nil
	close(client.snapshotDeleteChannel)
	client.snapshotDeleteChannel = nil
	close(client.snapshotRestoreChannel)
	client.snapshotRestoreChannel = nil
	close(client.snapshotsListChannel)
	client.snapshotsListChannel = nil
	close(client.cloudSnapshotUpdateChannel)
	client.cloudSnapshotUpdateChannel = nil
	close(client.cloudSnapshotDeleteChannel)
	client.cloudSnapshotDeleteChannel = nil
	close(client.quarkUpdateChannel)
	client.quarkUpdateChannel = nil
	close(client.quarkDeleteChannel)
	client.quarkDeleteChannel = nil
}

func (client *NVCClient) setupController(stopCh <-chan struct{}) error {
	informerFactory := quarkinformers.NewSharedInformerFactoryWithOptions(client.quarkClient, time.Second*30, quarkinformers.WithNamespace(client.nvcNamespace))
	volumeInformer := informerFactory.Quark().V1alpha1().NetAppVolumes()
	client.volumesLister = volumeInformer.Lister()
	client.volumesSynced = volumeInformer.Informer().HasSynced
	//client.workqueue = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumes")

	volumeInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		DeleteFunc: func(obj interface{}) {
			volume, ok := obj.(*quarkv1alpha1.NetAppVolume)
			if ok {
				client.handleVolumeDeletes(volume)
			}
		},
		UpdateFunc: func(old, new interface{}) {
			newVolume := new.(*quarkv1alpha1.NetAppVolume)
			oldVolume := old.(*quarkv1alpha1.NetAppVolume)

			// Filter previously seen NetAppVolumes.
			//key, err := cache.MetaNamespaceKeyFunc(new)

			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newVolume.ResourceVersion == oldVolume.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"volume":       newVolume.Name,
				"newRV":        newVolume.ResourceVersion,
				"oldRV":        oldVolume.ResourceVersion,
				"newVolStatus": newVolume.Status,
				"oldVolStatus": oldVolume.Status,
			}).Debug("NetAppVolume is updated.")

			//client.enqueueNetAppVolume(newVolume) //TODO: DEL
			//TODO: Ideally, a scheme like the way Trident diffs the old and
			// new objects make more sense here:
			//TODO: add support for multithreading
			client.handleVolumeUpdates(oldVolume, newVolume)
		},
	})

	snapshotInformer := informerFactory.Quark().V1alpha1().NetAppVolumeSnapshots()
	client.snapshotsLister = snapshotInformer.Lister()
	client.snapshotsSynced = snapshotInformer.Informer().HasSynced
	//client.workqueueSnapshot = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumeSnapshots")
	snapshotInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		UpdateFunc: func(old, new interface{}) {
			newVolSnap := new.(*quarkv1alpha1.NetAppVolumeSnapshot)
			oldVolSnap := old.(*quarkv1alpha1.NetAppVolumeSnapshot)

			// Filter previously seen NetAppVolumeSnapshots.
			//key, err := cache.MetaNamespaceKeyFunc(new)
			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newVolSnap.ResourceVersion == oldVolSnap.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"snapshot":         newVolSnap.Name,
				"newRV":            newVolSnap.ResourceVersion,
				"oldRV":            oldVolSnap.ResourceVersion,
				"newVolSnapStatus": newVolSnap.Status,
				"oldVolSnapStatus": oldVolSnap.Status,
			}).Debug("NetAppVolumeSnapshot is updated.")

			//TODO: add support for multithreading
			client.handleSnapshotUpdates(oldVolSnap, newVolSnap)
		},
	})

	snapshotsListInformer := informerFactory.Quark().V1alpha1().NetAppVolumeSnapshotsLists()
	client.snapshotsListLister = snapshotsListInformer.Lister()
	client.snapshotsListSynced = snapshotsListInformer.Informer().HasSynced
	//client.workqueueSnapshotsList = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumeSnapshotsList")
	snapshotsListInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		UpdateFunc: func(old, new interface{}) {
			newVolSnapLst := new.(*quarkv1alpha1.NetAppVolumeSnapshotsList)
			oldVolSnapLst := old.(*quarkv1alpha1.NetAppVolumeSnapshotsList)

			// Filter previously seen NetAppVolumeSnapshots.
			//key, err := cache.MetaNamespaceKeyFunc(new)
			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newVolSnapLst.ResourceVersion == oldVolSnapLst.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"snapshotsList":         newVolSnapLst.Name,
				"newRV":                 newVolSnapLst.ResourceVersion,
				"oldRV":                 oldVolSnapLst.ResourceVersion,
				"newVolSnapsListStatus": newVolSnapLst.Status,
				"oldVolSnapsListStatus": oldVolSnapLst.Status,
			}).Debug("NetAppVolumeSnapshotsList is updated.")

			//TODO: add support for multithreading
			client.handleSnapshotsListUpdates(oldVolSnapLst, newVolSnapLst)
			client.handleSnapshotRestoreUpdates(oldVolSnapLst, newVolSnapLst)
		},
	})

	cloudSnapshotInformer := informerFactory.Quark().V1alpha1().NetAppVolumeCloudSnapshots()
	client.cloudSnapshotsLister = cloudSnapshotInformer.Lister()
	client.cloudSnapshotsSynced = cloudSnapshotInformer.Informer().HasSynced
	//client.workqueuecloudSnapshot = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumeCloudSnapshot")
	cloudSnapshotInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		DeleteFunc: func(obj interface{}) {
			cloudSnapshot, ok := obj.(*quarkv1alpha1.NetAppVolumeCloudSnapshot)
			if ok {
				client.handleCloudSnapshotDeletes(cloudSnapshot)
			}
		},
		UpdateFunc: func(old, new interface{}) {
			newCloudSnap := new.(*quarkv1alpha1.NetAppVolumeCloudSnapshot)
			oldCloudSnap := old.(*quarkv1alpha1.NetAppVolumeCloudSnapshot)

			// Filter previously seen NetAppVolumeCloudSnapshots.
			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newCloudSnap.ResourceVersion == oldCloudSnap.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"cloudSnapshot":      newCloudSnap.Name,
				"newRV":              newCloudSnap.ResourceVersion,
				"oldRV":              oldCloudSnap.ResourceVersion,
				"newCloudSnapStatus": newCloudSnap.Status,
				"oldCloudSnapStatus": oldCloudSnap.Status,
			}).Debug("NetAppVolumeCloudSnapshot is updated.")

			//TODO: add support for multithreading
			client.handleCloudSnapshotUpdates(oldCloudSnap, newCloudSnap)
		},
	})

	quarkInformer := informerFactory.Quark().V1alpha1().Quarks()
	client.quarksLister = quarkInformer.Lister()
	client.quarksSynced = quarkInformer.Informer().HasSynced

	quarkInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		DeleteFunc: func(obj interface{}) {
			quark, ok := obj.(*quarkv1alpha1.Quark)
			if ok {
				client.handleQuarkDeletes(quark)
			}
		},
		UpdateFunc: func(old, new interface{}) {
			newQuark := new.(*quarkv1alpha1.Quark)
			oldQuark := old.(*quarkv1alpha1.Quark)

			// Filter previously seen Quark.
			_, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}
			if newQuark.ResourceVersion == oldQuark.ResourceVersion {
				return
			}
			log.WithFields(log.Fields{
				"quark":          newQuark.Name,
				"newRV":          newQuark.ResourceVersion,
				"oldRV":          oldQuark.ResourceVersion,
				"newQuarkStatus": newQuark.Status,
				"oldQuarkStatus": oldQuark.Status,
			}).Debug("Quark is updated.")

			//TODO: add support for multithreading
			client.handleQuarkUpdates(oldQuark, newQuark)
		},
	})

	informerFactory.Start(stopCh)

	log.Debug("Waiting for cache sync.")
	if ok := cache.WaitForCacheSync(stopCh, client.volumesSynced, client.snapshotsSynced, client.snapshotsListSynced, client.cloudSnapshotsSynced, client.quarksSynced); !ok {
		return fmt.Errorf("failed to wait for caches to sync")
	}

	log.Debug("Controller finished cache sync.")
	return nil
}

func (client *NVCClient) handleSnapshotRestoreUpdates(oldVolSnapList *quarkv1alpha1.NetAppVolumeSnapshotsList, newVolSnapList *quarkv1alpha1.NetAppVolumeSnapshotsList) {
	// send an event when SnapshotRestoreError condition added to SnapshotsList conditions
	if len(newVolSnapList.Status.Conditions) != len(oldVolSnapList.Status.Conditions) {
		for _, cond := range newVolSnapList.Status.Conditions {
			if cond.Type == quarkv1alpha1.SnapshotRestoreError && cond.Status == corev1.ConditionTrue {
				client.snapshotRestoreChannel <- &NetAppVolumeSnapshotRestore{
					VolumeUUID:       newVolSnapList.Name,
					ErrorCondition:   cond.Message,
					RestoredSnapshot: false,
				}
				log.WithFields(log.Fields{
					"snapshot":      newVolSnapList.Name,
					"oldConditions": oldVolSnapList.Status.Conditions,
					"newConditions": newVolSnapList.Status.Conditions,
				}).Debug("NetAppVolumeSnapshotsList conditions have been updated with SnapshotRestoreError condition.")
			}
		}
	}

	if newVolSnapList.Status.RestoredSnapshot != oldVolSnapList.Status.RestoredSnapshot {
		client.snapshotRestoreChannel <- &NetAppVolumeSnapshotRestore{
			VolumeUUID:       newVolSnapList.Name,
			ErrorCondition:   "",
			RestoredSnapshot: true,
		}
		log.WithFields(log.Fields{
			"snapshot":            newVolSnapList.Name,
			"oldRestoredSnapshot": oldVolSnapList.Status.RestoredSnapshot,
			"newRestoredSnapshot": newVolSnapList.Status.RestoredSnapshot,
		}).Debug("Detected SnapRestore.")
	}
}

func (client *NVCClient) handleVolumeDeletes(vol *quarkv1alpha1.NetAppVolume) {
	client.volumeDeleteChannel <- &NetAppVolumeDelete{vol}
}

func (client *NVCClient) handleVolumeUpdates(oldVol *quarkv1alpha1.NetAppVolume, newVol *quarkv1alpha1.NetAppVolume) {
	sendEvent := false

	// We send events when any of the NetAppVolume has converged to the desired state.
	_, newStatus, _ := client.NetAppVolumeUpdated(newVol)
	_, oldStatus, _ := client.NetAppVolumeUpdated(oldVol)
	if newStatus != oldStatus {
		log.WithFields(log.Fields{
			"volume":    newVol.Name,
			"newStatus": newStatus,
		}).Debug("NetApp Volume state has been updated.")
		sendEvent = true
	}
	// We send events when the NetAppVolume's conditions have changed.
	if len(newVol.Status.Conditions) != len(oldVol.Status.Conditions) {
		sendEvent = true
		log.WithFields(log.Fields{
			"volume":     newVol.Name,
			"conditions": newVol.Status.Conditions,
		}).Debug("NetApp Volume has updated conditions.")
	} else {
		for i, cond := range newVol.Status.Conditions {
			if oldVol.Status.Conditions[i].Type != cond.Type || oldVol.Status.Conditions[i].Status != cond.Status {
				sendEvent = true
				log.WithFields(log.Fields{
					"volume":     newVol.Name,
					"conditions": newVol.Status.Conditions,
				}).Debug("NetApp Volume has updated conditions.")
				break
			}
		}
	}

	if sendEvent {
		client.volumeUpdateChannel <- &NetAppVolumeUpdate{
			Volume:            newVol,
			Status:            newStatus,
			UpdatedConditions: newVol.Status.Conditions,
		}
	}
}

func (client *NVCClient) handleSnapshotUpdates(oldVolSnap *quarkv1alpha1.NetAppVolumeSnapshot, newVolSnap *quarkv1alpha1.NetAppVolumeSnapshot) {
	sendEvent := false

	// We send events when the NetAppVolumeSnapshot's conditions have changed.
	if len(newVolSnap.Status.Conditions) != len(oldVolSnap.Status.Conditions) {
		sendEvent = true
		log.WithFields(log.Fields{
			"snapshot":   newVolSnap.Name,
			"conditions": newVolSnap.Status.Conditions,
		}).Debug("NetApp Volume Snapshot has updated conditions.")
	} else {
		for i, cond := range newVolSnap.Status.Conditions {
			if oldVolSnap.Status.Conditions[i].Type != cond.Type || oldVolSnap.Status.Conditions[i].Status != cond.Status {
				sendEvent = true
				log.WithFields(log.Fields{
					"snapshot":   newVolSnap.Name,
					"conditions": newVolSnap.Status.Conditions,
				}).Debug("NetApp Volume Snapshot has updated conditions.")
				break
			}
		}
	}

	if sendEvent {
		client.snapshotUpdateChannel <- &NetAppVolumeSnapshotUpdate{
			Snapshot:          newVolSnap,
			UpdatedConditions: newVolSnap.Status.Conditions,
		}
	}
}

// Difference between two snapshotsList. All the snapshots in the oldList that are not in the newList
func differenceSnapshots(oldList, newList []quarkv1alpha1.NetAppVolumeSnapshotInfo) (diff []string) {
	listMap := make(map[string]int)

	key := ""
	for _, snap := range newList {
		key = snap.Name + snap.CreationTime.String()
		listMap[key] = listMap[key] + 1
	}

	for _, snap := range oldList {
		key = snap.Name + snap.CreationTime.String()
		if count, ok := listMap[key]; !ok || count == 0 {
			diff = append(diff, snap.Name)
		} else {
			listMap[key] = listMap[key] - 1
		}
	}
	return
}

func (client *NVCClient) snapshotsListConditionsUpdated(oldVolSnap *quarkv1alpha1.NetAppVolumeSnapshotsList, newVolSnap *quarkv1alpha1.NetAppVolumeSnapshotsList) bool {
	if len(newVolSnap.Status.Conditions) != len(oldVolSnap.Status.Conditions) {
		return true
	} else {
		var (
			newConditionLastTransitionTime metav1.Time
			oldConditionLastTransitionTime metav1.Time
		)
		for _, cond := range newVolSnap.Status.Conditions {
			if cond.Type == quarkv1alpha1.SnapshotsListReady && cond.Status == corev1.ConditionTrue {
				newConditionLastTransitionTime = cond.LastTransitionTime
			}
		}
		for _, cond := range oldVolSnap.Status.Conditions {
			if cond.Type == quarkv1alpha1.SnapshotsListReady && cond.Status == corev1.ConditionTrue {
				oldConditionLastTransitionTime = cond.LastTransitionTime
			}
		}
		newConditionUpdateTime := newConditionLastTransitionTime.Time
		oldConditionUpdateTime := oldConditionLastTransitionTime.Time
		log.WithFields(log.Fields{
			"snapshot":      newVolSnap.Name,
			"oldConditions": oldVolSnap.Status.Conditions,
			"newConditions": newVolSnap.Status.Conditions,
		}).Debug("NetApp Volume Snapshots condition has been updated.")

		return !newConditionUpdateTime.Equal(oldConditionUpdateTime)
	}

}

func (client *NVCClient) handleSnapshotsListUpdates(oldVolSnap *quarkv1alpha1.NetAppVolumeSnapshotsList, newVolSnap *quarkv1alpha1.NetAppVolumeSnapshotsList) {
	// We send events when the NetAppVolumeSnapshotsList's list have changed.
	if client.snapshotsListConditionsUpdated(oldVolSnap, newVolSnap) {
		client.snapshotsListChannel <- &NetAppVolumeSnapshotsListUpdate{
			VolumeUUID:           newVolSnap.Name,
			SnapshotsListUpdated: true,
		}
		log.WithFields(log.Fields{
			"volume":    newVolSnap.Name,
			"snapshots": newVolSnap.Status.Snapshots,
		}).Debug("NetAppVolumeSnapshotsList has been updated.")
	}
	// All the snapshots that were in snapshots old list before but not in new list are deleted.
	// We are comparing Snapshots instead of deletedSnapshots since multiple updates can be coalesced together at client-go.
	deletedList := differenceSnapshots(oldVolSnap.Status.Snapshots, newVolSnap.Status.Snapshots)
	if len(deletedList) > 0 {
		client.snapshotDeleteChannel <- &NetAppVolumeSnapshotDelete{
			VolumeUUID:       newVolSnap.Name,
			DeletedSnapshots: deletedList,
		}
		log.WithFields(log.Fields{
			"volume":    newVolSnap.Name,
			"snapshots": newVolSnap.Status.DeletedSnapshots,
		}).Debug("Detected deletion of snapshots.")
	}
}

func (client *NVCClient) handleCloudSnapshotDeletes(cloudSnap *quarkv1alpha1.NetAppVolumeCloudSnapshot) {
	client.cloudSnapshotDeleteChannel <- &NetAppVolumeCloudSnapshotDelete{cloudSnap}
}

func (client *NVCClient) handleCloudSnapshotUpdates(oldCloudSnap *quarkv1alpha1.NetAppVolumeCloudSnapshot, newCloudSnap *quarkv1alpha1.NetAppVolumeCloudSnapshot) {
	sendEvent := false

	oldCloudSnapInfo := getCloudSnapshotInfo(oldCloudSnap)
	newCloudSnapInfo := getCloudSnapshotInfo(newCloudSnap)

	if oldCloudSnapInfo.Ready != newCloudSnapInfo.Ready {
		sendEvent = true
	}
	if oldCloudSnapInfo.Uploading != newCloudSnapInfo.Uploading {
		sendEvent = true
	}
	if oldCloudSnapInfo.Deleting != newCloudSnapInfo.Deleting {
		sendEvent = true
	}
	if oldCloudSnapInfo.EndpointUUID != newCloudSnapInfo.EndpointUUID {
		sendEvent = true
	}
	if oldCloudSnapInfo.LogicalSize != newCloudSnapInfo.LogicalSize {
		sendEvent = true
	}
	if oldCloudSnapInfo.CompletionPercent != newCloudSnapInfo.CompletionPercent {
		sendEvent = true
	}
	if oldCloudSnapInfo.Error != newCloudSnapInfo.Error {
		if (oldCloudSnapInfo.Error != nil) &&
			(newCloudSnapInfo.Error != nil) {
			if oldCloudSnapInfo.Error.Error() != newCloudSnapInfo.Error.Error() {
				sendEvent = true
			}
		} else {
			sendEvent = true
		}
	}
	if oldCloudSnapInfo.CreationTime != newCloudSnapInfo.CreationTime {
		if (oldCloudSnapInfo.CreationTime != nil) &&
			(newCloudSnapInfo.CreationTime != nil) {
			if *oldCloudSnapInfo.CreationTime != *newCloudSnapInfo.CreationTime {
				sendEvent = true
			}
		} else {
			sendEvent = true
		}
	}

	if sendEvent {
		client.cloudSnapshotUpdateChannel <- &NetAppVolumeCloudSnapshotUpdate{
			CloudSnapshot: newCloudSnap,
			Status:        newCloudSnapInfo,
		}
	}
}

func (client *NVCClient) handleQuarkUpdates(oldQuark *quarkv1alpha1.Quark, newQuark *quarkv1alpha1.Quark) {
	sendEvent := false

	oldNtapQuarkStatus := getNetAppQuarkStatus(oldQuark)
	newNtapQuarkStatus := getNetAppQuarkStatus(newQuark)

	if oldNtapQuarkStatus.OperatorReady != newNtapQuarkStatus.OperatorReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.VolumeControllerReady != newNtapQuarkStatus.VolumeControllerReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.HighAvailabilityReady != newNtapQuarkStatus.HighAvailabilityReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.ExternalVolumeAccessReady != newNtapQuarkStatus.ExternalVolumeAccessReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.CloudsnapshotCapabilityReady != newNtapQuarkStatus.CloudsnapshotCapabilityReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.IPDReady != newNtapQuarkStatus.IPDReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.EtcdClusterReady != newNtapQuarkStatus.EtcdClusterReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.NodevolReady != newNtapQuarkStatus.NodevolReady {
		sendEvent = true
	}

	if oldNtapQuarkStatus.Version != newNtapQuarkStatus.Version {
		sendEvent = true
	}

	if sendEvent {
		client.quarkUpdateChannel <- &NetAppQuarkUpdate{
			Quark:  newQuark,
			Status: newNtapQuarkStatus,
		}
	}
}

func (client *NVCClient) handleQuarkDeletes(quark *quarkv1alpha1.Quark) {
	client.quarkDeleteChannel <- &NetAppQuarkDelete{Quark: quark}
}

// ConstructNetAppVolumeByRef - Constructs a new newapp volume by passing in references to export & snapshot policy objects
func ConstructNetAppVolumeByRef(ownerID, displayName, sizeString, volumeUUID, volumePath string, serviceLevel ServiceLevel, exportPolicy *quarkv1alpha1.NetAppExportPolicy, snapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy) (volume *quarkv1alpha1.NetAppVolume, err error) {
	return ConstructNetAppVolume(ownerID, displayName, sizeString, volumeUUID, volumePath, exportPolicy.Name, snapshotPolicy.Name, serviceLevel)
}

// Constructs NetAppVolume definition.
// Constructs NetAppVolume definition helper routine.
// Sets the common field for NetAppVolume
// ownerID: OwnerID of the volume.
// dispalyName: Name of the volume provided by customer.
// sizeString: Volume Size.
// volumeUUID: Unique ID of volume in k8s cluster.
// volumePath: Volume mount path.
// serviceLevel: Standard for alpha.
// snapshotPolicy: Snapshot Policy object
func constructBaseNetAppVolume(ownerID, displayName, sizeString, volumeUUID, volumePath, snapshotPolicyName string, serviceLevel ServiceLevel) (volume *quarkv1alpha1.NetAppVolume, err error) {
	defer func() {
		if e := recover(); e != nil {
			err = fmt.Errorf("invalid value for size")
		}
	}()

	// Validate service level
	switch serviceLevel {
	case ServiceLevelStandard, ServiceLevelRegionalStandard:
	default:
		return nil, fmt.Errorf("invalid value for service level")
	}

	// validate UUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for UUID")
	}
	// Validate size
	size, err := resource.ParseQuantity(sizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}
	var minVolSizeDifference = size.Value() - minVolSize.Value()
	var maxVolSizeDifference = size.Value() - maxVolSize.Value()
	if minVolSizeDifference < 0 || maxVolSizeDifference > 0 || minVolSizeDifference%incrementalSize.Value() != 0 {
		return nil, fmt.Errorf("size=%s, volume size should be a minimum of %s, maximum of %s and a multiple of %s", size.String(), minVolSize.String(), maxVolSize.String(), incrementalSize.String())
	}

	// Validate ownerID
	if ownerID == "" {
		return nil, fmt.Errorf("volume OwnerID required")
	}

	volume = &quarkv1alpha1.NetAppVolume{
		metav1.TypeMeta{},
		metav1.ObjectMeta{
			Name: volumeUUID,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
		},
		quarkv1alpha1.NetAppVolumeSpec{
			Size:           size,
			DisplayName:    displayName,
			VolumePath:     volumePath,
			ServiceLevel:   string(serviceLevel),
			SnapshotPolicy: snapshotPolicyName,
		},
		quarkv1alpha1.NetAppVolumeStatus{},
	}
	return
}

// Constructs NetAppVolume definition.
// ownerID: OwnerID of the volume.
// dispalyName: Name of the volume provided by customer.
// sizeString: Volume Size.
// volumeUUID: Unique ID of volume in k8s cluster.
// volumePath: Volume mount path.
// serviceLevel: Standard for alpha.
// exportPolicy: Export Policy object
// snapshotPolicy: Snapshot Policy object
func ConstructNetAppVolume(ownerID, displayName, sizeString, volumeUUID, volumePath, exportPolicyName, snapshotPolicyName string, serviceLevel ServiceLevel) (volume *quarkv1alpha1.NetAppVolume, err error) {
	err = nvc.ValidateNFSConfiguration(volumePath)
	if err != nil {
		return nil, err
	}
	volume, err = constructBaseNetAppVolume(ownerID, displayName, sizeString, volumeUUID, volumePath, snapshotPolicyName, serviceLevel)
	if err != nil {
		return nil, err
	}
	volume.Spec.ExportPolicy = exportPolicyName
	volume.Spec.Protocol = "nfs3"

	return volume, err
}

func (client *NVCClient) NetAppVolumeNoSnapDirUpdate(vol *quarkv1alpha1.NetAppVolume, val bool) (volume *quarkv1alpha1.NetAppVolume, err error) {
	vol.SetNoSnapDir(val)
	newVol, err := client.netAppVolumeUpdate(vol)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return newVol, err

}

func (client *NVCClient) NetAppVolumeSnapshotPolicyUpdate(vol *quarkv1alpha1.NetAppVolume, snapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy) (volume *quarkv1alpha1.NetAppVolume, err error) {
	if snapshotPolicy == nil {
		return nil, fmt.Errorf("snapshot policy can not be nil")
	}
	vol.Spec.SnapshotPolicy = snapshotPolicy.Name
	newVol, err := client.netAppVolumeUpdate(vol)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return newVol, err
}

func (client *NVCClient) NetAppVolumeExportPolicyUpdate(vol *quarkv1alpha1.NetAppVolume, exportPolicy *quarkv1alpha1.NetAppExportPolicy) (volume *quarkv1alpha1.NetAppVolume, err error) {
	if exportPolicy == nil {
		return nil, fmt.Errorf("export policy can not be nil")
	}
	vol.Spec.ExportPolicy = exportPolicy.Name
	newVol, err := client.netAppVolumeUpdate(vol)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return newVol, err
}

func (client *NVCClient) NetAppVolumeSizeUpdate(vol *quarkv1alpha1.NetAppVolume, newSizeString string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	size, err := resource.ParseQuantity(newSizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}
	var minVolSizeDifference = size.Value() - minVolSize.Value()
	var maxVolSizeDifference = size.Value() - maxVolSize.Value()
	if minVolSizeDifference < 0 || maxVolSizeDifference > 0 || minVolSizeDifference%incrementalSize.Value() != 0 {
		return nil, fmt.Errorf("size=%s, volume size should be a minimum of %s, maximum of %s and a multiple of %s", size.String(), minVolSize.String(), maxVolSize.String(), incrementalSize.String())
	}

	vol.Spec.Size = size

	newVol, err := client.netAppVolumeUpdate(vol)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return newVol, err
}

// Update EncryptionEnabled field for SMB volume
func (client *NVCClient) NetAppVolumeSMBEncryptionUpdate(vol *quarkv1alpha1.NetAppVolume, encryptionEnabled bool) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.NetAppVolumeGet(vol.Name)
	if err != nil {
		return nil, err
	}
	volume.Spec.SMBConfiguration.EncryptionEnabled = encryptionEnabled
	volume, err = client.netAppVolumeUpdate(volume)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return
}

// Abort NetAppVolumeSMBEncryptionUpdate() operation
func (client *NVCClient) NetAppVolumeSMBEncryptionUpdateAbort(vol *quarkv1alpha1.NetAppVolume) (volume *quarkv1alpha1.NetAppVolume, err error) {
	return client.NetAppVolumeSMBEncryptionUpdate(vol, vol.Status.SMBConfiguration.EncryptionEnabled)
}

// Reset Machine Account password for SMB volume
func (client *NVCClient) NetAppVolumeSMBPasswordReset(volumeUUID string, user, password string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return nil, err
	}
	secretName := util.GetVolPwdResetADSecretName(volume.Name)
	_, err = client.createOrUpdateADSecretWithOwnerRef(volume, secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": volume.Name,
			"secret": secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	volume.Spec.SMBPasswordResetCount = volume.Spec.SMBPasswordResetCount + 1
	volume, err = client.netAppVolumeUpdate(volume)
	if err != nil {
		delErr := client.SecretDelete(secretName)
		if delErr != nil {
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"secret": secretName,
				"error":  delErr,
			}).Error("Failed to delete secret.")
		}
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return
}

// Abort machine-account password reset operation for SMB volume.
// SDE should abort the password reset operation if it times out.
// Abort will delete the secret and avoid retrying operation.
func (client *NVCClient) NetAppVolumeSMBPasswordResetAbort(volumeUUID string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	secretName := util.GetVolPwdResetADSecretName(volumeUUID)
	err = client.SecretDelete(secretName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, err
	}
	volume, err = client.NetAppVolumeGet(volumeUUID)
	if err != nil {
		return nil, err
	}
	volume.Spec.SMBPasswordResetCount = volume.Status.SMBPasswordResetCount
	volume, err = client.netAppVolumeUpdate(volume)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return volume, err
}

// Update DNS IP for SMB volume
func (client *NVCClient) netAppVolumeSMBDNSUpdate(vol *quarkv1alpha1.NetAppVolume, dnsIPs []string) (volume *quarkv1alpha1.NetAppVolume, err error) {
	volume, err = client.NetAppVolumeGet(vol.Name)
	if err != nil {
		return nil, err
	}
	volume.Spec.SMBConfiguration.DNSIPs = dnsIPs
	volume, err = client.netAppVolumeUpdate(volume)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return
}

type ExportRule quarkv1alpha1.ExportRule

// Construct NetAppExportPolicy definition
// name: Name should be unique. Suggested Naming convention is volName-revisionNumber.
// rules: ExportPolicy rule
func ConstructNetAppExportPolicy(name string, rules []ExportRule, volumeUUID string, ownerID string) (policy *quarkv1alpha1.NetAppExportPolicy, err error) {
	validProtocols := []string{"nfs3", "nfs4", "cifs"}
	validSecurityType := []string{"unix", "sys", "any", "never"}
	policy = &quarkv1alpha1.NetAppExportPolicy{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: name,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppExportPolicySpec{
			Rules: []quarkv1alpha1.ExportRule{},
		},
	}
	var policyRule quarkv1alpha1.ExportRule
	for i, rl := range rules {
		// TODO: Validate NetAppExpolicy Name if we are following any naming convention

		if !util.StringSetContains(validSecurityType, rl.SecurityType) {
			return nil, fmt.Errorf("exportPolicy: %s: invalid security type %s for rule %d", name, rl.SecurityType, i)
		}
		if !util.StringSetContains(validProtocols, rl.Protocol) {
			return nil, fmt.Errorf("exportPolicy: %s: invalid protocol %s for rule %d", name, rl.Protocol, i)
		}

		if rl.AccessMode != "rw" && rl.AccessMode != "ro" {
			return nil, fmt.Errorf("exportPolicy: %s: invalid access mode %s for rule %d", name, rl.AccessMode, i)
		}
		policyRule = quarkv1alpha1.ExportRule{
			RuleIndex:    rl.RuleIndex,
			Clients:      rl.Clients,
			AccessMode:   rl.AccessMode,
			SecurityType: rl.SecurityType,
			Protocol:     rl.Protocol,
		}
		policy.Spec.Rules = append(policy.Spec.Rules, policyRule)
	}
	return
}

type CronSchedule quarkv1alpha1.CronSchedule

func validateTimeInput(policyName string, arr []int64, maxItems int, minVal int64, maxVal int64, item string) (err error) {
	if len(arr) > maxItems {
		return fmt.Errorf("snapshotPolicy: %s: more than %d items in the %s list not allowed", policyName, maxItems, item)
	}
	keys := make(map[int64]bool)
	for _, val := range arr {
		if _, value := keys[val]; !value {
			keys[val] = true
		} else {
			return fmt.Errorf("snapshotPolicy: %s: duplicate value in an array %s", policyName, item)
		}
		if val < minVal || val > maxVal {
			return fmt.Errorf("snapshotPolicy: %s: %d is an invalid value for %s, Valid values are %d-%d",
				policyName, val, item, minVal, maxVal)
		}
	}
	return nil
}

// Construct Netapp Snapshot Policy definition
// DaysOfWeek: Format Sunday -0,... ,Saturday - 6
func ConstructNetAppSnapshotPolicy(name string, isEnabled bool, schedules []CronSchedule, volumeUUID string, ownerID string) (policy *quarkv1alpha1.NetAppSnapshotPolicy, err error) {
	policy = &quarkv1alpha1.NetAppSnapshotPolicy{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: name,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppSnapshotPolicySpec{
			Schedules: []quarkv1alpha1.CronSchedule{},
		},
	}

	policy.Spec.Enabled = isEnabled
	var policySchedule quarkv1alpha1.CronSchedule
	for _, sc := range schedules {
		if err = validateTimeInput(name, sc.Months, 12, 1, 12, "Months"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(name, sc.DaysOfMonth, 31, 1, 31, "DaysOfMonth"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(name, sc.DaysOfWeek, 7, 0, 6, "DaysOfWeek"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(name, sc.Hours, 24, 0, 23, "Hours"); err != nil {
			return nil, err
		}
		if err = validateTimeInput(name, sc.Minutes, 60, 0, 59, "Minutes"); err != nil {
			return nil, err
		}

		policySchedule = quarkv1alpha1.CronSchedule{
			Prefix:          sc.Prefix,
			Months:          sc.Months,
			DaysOfMonth:     sc.DaysOfMonth,
			DaysOfWeek:      sc.DaysOfWeek,
			Hours:           sc.Hours,
			Minutes:         sc.Minutes,
			SnapshotsToKeep: sc.SnapshotsToKeep,
		}
		policy.Spec.Schedules = append(policy.Spec.Schedules, policySchedule)
	}
	return
}

// Construct Netapp Volume Snapshot definition
func ConstructNetAppVolumeSnapshot(ownerID, snapUUID, volumeUUID, displayName string) (volumeSnapshot *quarkv1alpha1.NetAppVolumeSnapshot, err error) {
	// validate volumeUUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for volumeUUID")
	}
	// validate snapUUID
	_, err = uuid.FromString(snapUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for snapUUID")
	}
	// validate displayName
	prefixes := []string{"hourly", "daily", "monthly", "weekly", "cloud-snapshot"}
	for _, prefix := range prefixes {
		if strings.HasPrefix(strings.ToLower(displayName), prefix) {
			return nil, fmt.Errorf("invalid prefix: \"%s\". Snapshot display name must not have prefixes: %v", prefix, prefixes)
		}
	}

	volumeSnapshot = &quarkv1alpha1.NetAppVolumeSnapshot{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: snapUUID,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeSnapshotSpec{
			DisplayName: displayName,
		},
	}
	return
}

// ConfigureDNS configures StubDomains in KubeDNS configmap
// domainName: Domain Name
// ipaddresses: list of DNS ipaddresses
func (client *NVCClient) ConfigureStubDomains(domainName string, ipaddresses []string) (err error) {
	cm, err := client.kubeClient.CoreV1().ConfigMaps("kube-system").Get(
		context.TODO(), "kube-dns", metav1.GetOptions{})
	if err != nil {
		return err
	}

	stubDomains := make(map[string][]string)

	var (
		newData = []byte{}
	)
	if _, ok := cm.Data["stubDomains"]; ok {
		oldData := []byte(cm.Data["stubDomains"])
		if err = json.Unmarshal([]byte(oldData), &stubDomains); err != nil {
			log.WithFields(log.Fields{
				"domainName": domainName,
				"ipaddress":  ipaddresses,
				"oldData":    oldData,
			}).Error("Error unmarshalling oldData.")
			return fmt.Errorf("error configuring KubeDNS domain %s: %s", domainName, err)
		}
		if _, ok := stubDomains[domainName]; ok {
			log.WithFields(log.Fields{
				"domainName": domainName,
				"ipaddress":  ipaddresses,
			}).Info("Replacing DNS IP address in KubeDNS configmap.")
		} else {
			log.WithFields(log.Fields{
				"domainName": domainName,
				"ipaddress":  ipaddresses,
			}).Info("Adding StubDomain to KubeDNS configmap.")
		}
	} else {
		log.WithFields(log.Fields{
			"domainName": domainName,
			"ipaddress":  ipaddresses,
		}).Info("Adding first StubDomain to KubeDNS configmap.")
	}

	stubDomains[domainName] = ipaddresses

	newData, err = json.Marshal(stubDomains)
	if err != nil {
		log.WithFields(log.Fields{
			"domainName": domainName,
			"ipaddress":  ipaddresses,
		}).Error("Error marshalling newData.")
		return fmt.Errorf("error configuring KubeDNS domain %s: %s", domainName, err)
	}

	if cm.Data == nil {
		cm.Data = make(map[string]string)
	}

	cm.Data["stubDomains"] = string(newData)

	_, err = client.kubeClient.CoreV1().ConfigMaps("kube-system").Update(
		context.TODO(), cm, metav1.UpdateOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"domainName": domainName,
			"ipaddress":  ipaddresses,
		}).Error("Error configuring KubeDNS.")
		return fmt.Errorf("error configuring KubeDNS domain %s: %s", domainName, err)
	}
	return err
}

// RemoveDNS removes StubDomain entry from KubeDNS configmap
// domainName: Domain name
func (client *NVCClient) RemoveStubDomains(domainName string) (err error) {
	cm, err := client.kubeClient.CoreV1().ConfigMaps("kube-system").Get(
		context.TODO(), "kube-dns", metav1.GetOptions{})
	if err != nil {
		return err
	}

	stubDomains := make(map[string][]string)

	var (
		newData = []byte{}
		oldData = []byte(`{}`)
	)
	if _, ok := cm.Data["stubDomains"]; !ok {
		return nil
	}
	oldData = []byte(cm.Data["stubDomains"])

	if err = json.Unmarshal([]byte(oldData), &stubDomains); err != nil {
		log.WithFields(log.Fields{
			"domainName": domainName,
			"oldData":    oldData,
		}).Error("Error unmarshalling oldData.")
		return fmt.Errorf("error removing KubeDNS domain %s: %s", domainName, err)
	}
	if _, ok := stubDomains[domainName]; !ok {
		return nil
	}
	delete(stubDomains, domainName)
	if len(stubDomains) == 0 {
		delete(cm.Data, "stubDomains")
	} else {
		newData, err = json.Marshal(stubDomains)
		if err != nil {
			log.WithFields(log.Fields{
				"domainName": domainName,
			}).Error("Error marshalling newData.")
			return fmt.Errorf("error removing KubeDNS domain %s: %s", domainName, err)
		}
		cm.Data["stubDomains"] = string(newData)
	}

	_, err = client.kubeClient.CoreV1().ConfigMaps("kube-system").Update(
		context.TODO(), cm, metav1.UpdateOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"domainName": domainName,
		}).Error("Error updating kube-dns configmap.")
		return fmt.Errorf("error removing KubeDNS domain %s: %s", domainName, err)
	}
	return err
}

func ConstructNetAppVolumeCloudSnapshot(ownerID, cloudSnapshotUUID, volumeUUID, displayName, container, server string, port int32, secretName, secretNamespace string) (volumeCloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot, err error) {
	// validate volumeUUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for volumeUUID")
	}
	// validate cloudSnapshotUUID
	_, err = uuid.FromString(cloudSnapshotUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for cloudSnapshotUUID")
	}

	volumeCloudSnapshot = &quarkv1alpha1.NetAppVolumeCloudSnapshot{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: cloudSnapshotUUID,
			Labels: map[string]string{
				quarkconfig.VolumeLabel:  volumeUUID,
				quarkconfig.OwnerIDLabel: ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppVolumeCloudSnapshotSpec{
			DisplayName: displayName,
			Imported:    false,
			Container:   container,
			Server:      server,
			Port:        port,
			SecretRef: corev1.SecretReference{
				Name:      secretName,
				Namespace: secretNamespace,
			},
		},
	}
	return
}

func (client *NVCClient) NetAppVolumeCloudSnapshotCreate(volumeCloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) (*quarkv1alpha1.NetAppVolumeCloudSnapshot, error) {
	metaLabels := volumeCloudSnapshot.ObjectMeta.GetLabels()
	volumeUUID := metaLabels[quarkconfig.VolumeLabel]
	_, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).Get(
		context.TODO(), volumeUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	if volumeCloudSnapshot.Spec.Imported {
		return nil, fmt.Errorf("cannot create imported cloud snapshot")
	}
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).Create(
		context.TODO(), volumeCloudSnapshot, metav1.CreateOptions{})
}

func (client *NVCClient) getCloudSnapshot(cloudSnapshotUUID string) (*quarkv1alpha1.NetAppVolumeCloudSnapshot, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).Get(
		context.TODO(), cloudSnapshotUUID, metav1.GetOptions{})
}

func cloudSnapshotReady(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) bool {
	if cloudSnapshot.Spec.Imported {
		return true
	}
	conditions := cloudSnapshot.Status.Conditions
	for i := range conditions {
		if conditions[i].Type == quarkv1alpha1.CloudSnapshotReady {
			return conditions[i].Status == corev1.ConditionTrue
		}
	}
	return false
}

func cloudSnapshotUploading(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) bool {
	if cloudSnapshot.Spec.Imported {
		return false
	}
	conditions := cloudSnapshot.Status.Conditions
	for i := range conditions {
		if conditions[i].Type == quarkv1alpha1.CloudSnapshotCreating {
			return conditions[i].Status == corev1.ConditionTrue
		}
	}
	return false
}

func cloudSnapshotDeleting(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) bool {
	return (cloudSnapshot.ObjectMeta.DeletionTimestamp != nil)
}

func cloudSnapshotError(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) error {
	conditions := cloudSnapshot.Status.Conditions
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.CloudSnapshotError &&
			cond.Status == corev1.ConditionTrue {
			return fmt.Errorf(cond.Message)
		}
	}
	return nil
}

func (client *NVCClient) NetAppVolumeCloudSnapshotReady(cloudSnapshotUUID string) bool {
	cloudSnapshotInfo, err := client.NetAppVolumeCloudSnapshotGet(cloudSnapshotUUID)
	if err != nil {
		return false
	}

	return cloudSnapshotInfo.Ready
}

func ConstructNetAppVolumeCloudSnapshotImport(ownerID, cloudSnapshotUUID, volumeUUID, displayName, container, server string, port int32, secretName, secretNamespace, endpointUUID string, logicalSize int64, creationTime time.Time) (volumeCloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot, err error) {
	// validate endpointUUID
	if endpointUUID == "" {
		return nil, fmt.Errorf("endpointUUID not provided")
	}

	volumeCloudSnapshot, err = ConstructNetAppVolumeCloudSnapshot(ownerID, cloudSnapshotUUID, volumeUUID, displayName, container, server, port, secretName, secretNamespace)
	if err != nil {
		return
	}

	volumeCloudSnapshot.Spec.Imported = true

	// XXX Client shouldn't be setting status. We'll handle populating the status in the C2C controller. This info can be placed in the spec or perhaps stored in the bucket for the controller to pull down.
	volumeCloudSnapshot.Status = quarkv1alpha1.NetAppVolumeCloudSnapshotStatus{
		EndpointUUID: endpointUUID,
		LogicalSize:  logicalSize,
		CreationTime: metav1.NewTime(creationTime),
	}
	return
}

func (client *NVCClient) NetAppVolumeCloudSnapshotImport(volumeCloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) (*quarkv1alpha1.NetAppVolumeCloudSnapshot, error) {
	if !volumeCloudSnapshot.Spec.Imported {
		return nil, fmt.Errorf("cannot import created cloud snapshot")
	}

	newCloudSnap, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).Create(
		context.TODO(), volumeCloudSnapshot, metav1.CreateOptions{})

	if err != nil {
		return nil, err
	}

	newCloudSnap.Status = volumeCloudSnapshot.Status

	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).UpdateStatus(
		context.TODO(), newCloudSnap, metav1.UpdateOptions{})
}

func (client *NVCClient) NetAppVolumeCloudSnapshotRestore(volume *quarkv1alpha1.NetAppVolume, cloudSnapshotUUID string) (*quarkv1alpha1.NetAppVolume, error) {

	cloudSnapshot, err := client.getCloudSnapshot(cloudSnapshotUUID)
	if err != nil {
		return nil, err
	}

	if !cloudSnapshotReady(cloudSnapshot) {
		return nil, fmt.Errorf("cannot restore, CloudSnapshot not ready: %s", cloudSnapshotUUID)
	}

	snapSize := cloudSnapshot.Status.LogicalSize
	volSize := volume.Spec.Size.Value()

	// Volume must be at least as large as the snapshot
	if volSize < snapSize {
		return nil, fmt.Errorf("cannot restore, volume size must be at least %d bytes", snapSize)
	}

	volume.Spec.CloudSnapshot = cloudSnapshotUUID

	return client.NetAppVolumeCreate(volume)
}

func (client *NVCClient) NetAppVolumeCloudSnapshotRestoreSMB(volume *quarkv1alpha1.NetAppVolume, cloudSnapshotUUID, user, password string) (*quarkv1alpha1.NetAppVolume, error) {

	cloudSnapshot, err := client.getCloudSnapshot(cloudSnapshotUUID)
	if err != nil {
		return nil, err
	}

	if !cloudSnapshotReady(cloudSnapshot) {
		return nil, fmt.Errorf("cannot restore, CloudSnapshot not ready: %s", cloudSnapshotUUID)
	}

	snapSize := cloudSnapshot.Status.LogicalSize
	volSize := volume.Spec.Size.Value()

	// Volume must be at least as large as the snapshot
	if volSize < snapSize {
		return nil, fmt.Errorf("cannot restore, volume size must be at least %d bytes", snapSize)
	}

	volume.Spec.CloudSnapshot = cloudSnapshotUUID

	return client.NetAppVolumeCreateSMB(volume, user, password)
}

// NetAppVolumeCloudSnapshotRestored returns "true" when the cloud snapshot is restored for a given volume.
func (client *NVCClient) NetAppVolumeCloudSnapshotRestored(volume *quarkv1alpha1.NetAppVolume) bool {

	cloudSnapUUID := volume.Spec.CloudSnapshot
	if cloudSnapUUID != "" {
		// check for cloud snap restored condition to be true
		for _, condition := range volume.Status.Conditions {
			if condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestored {
				return condition.Status == corev1.ConditionTrue
			}
		}
	}
	return false
}

func (client *NVCClient) NetAppVolumeCloudSnapshotDelete(cloudSnapshotUUID string) error {
	return client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).Delete(
		context.TODO(), cloudSnapshotUUID, metav1.DeleteOptions{})
}

func (client *NVCClient) NetAppVolumeCloudSnapshotDeleted(cloudSnapshotUUID string) bool {
	_, err := client.getCloudSnapshot(cloudSnapshotUUID)
	if err != nil {
		if errors.IsNotFound(err) {
			return true
		} else {
			log.WithFields(log.Fields{
				"error": err,
			}).Debug("getCloudSnapshot failed.")
		}
	}
	return false
}

type NetAppVolumeCloudSnapshotInfo struct {
	CloudSnapshotUUID string
	DisplayName       string
	OwnerID           string
	VolumeUUID        string
	Ready             bool
	Uploading         bool
	Deleting          bool
	EndpointUUID      string
	BytesTransferred  int64
	LogicalSize       int64
	CompletionPercent int64
	Error             error
	// Timestamps with location set to UTC.
	LastTransitionTime *time.Time
	LastHeartbeatTime  *time.Time
	CreationTime       *time.Time
}

func getCloudSnapshotInfo(cloudSnapshot *quarkv1alpha1.NetAppVolumeCloudSnapshot) NetAppVolumeCloudSnapshotInfo {
	cloudSnapInfo := NetAppVolumeCloudSnapshotInfo{
		CloudSnapshotUUID: cloudSnapshot.Name,
		DisplayName:       cloudSnapshot.Spec.DisplayName,
		OwnerID:           cloudSnapshot.Labels[quarkconfig.OwnerIDLabel],
		VolumeUUID:        cloudSnapshot.Labels[quarkconfig.VolumeLabel],
		Ready:             cloudSnapshotReady(cloudSnapshot),
		Uploading:         cloudSnapshotUploading(cloudSnapshot),
		Deleting:          cloudSnapshotDeleting(cloudSnapshot),
		EndpointUUID:      cloudSnapshot.Status.EndpointUUID,
		BytesTransferred:  cloudSnapshot.Status.BytesTransferred,
		LogicalSize:       cloudSnapshot.Status.LogicalSize,
		CompletionPercent: cloudSnapshot.Status.CompletionPercent,
		Error:             cloudSnapshotError(cloudSnapshot),
	}

	if len(cloudSnapshot.Status.Conditions) > 0 {
		// There should only be 0 or 1 conditions
		transitionUTC := cloudSnapshot.Status.Conditions[0].LastTransitionTime.UTC()
		cloudSnapInfo.LastTransitionTime = &transitionUTC
		heartbeatUTC := cloudSnapshot.Status.Conditions[0].LastHeartbeatTime.UTC()
		cloudSnapInfo.LastHeartbeatTime = &heartbeatUTC
	}

	if !cloudSnapshot.Status.CreationTime.IsZero() {
		creationUTC := cloudSnapshot.Status.CreationTime.UTC()
		cloudSnapInfo.CreationTime = &creationUTC
	}

	return cloudSnapInfo
}

func getNetAppQuarkStatus(quark *quarkv1alpha1.Quark) NetAppQuarkUpdateStatus {
	ntapQuarkStatus := NetAppQuarkUpdateStatus{}

	// version reconciled ?
	quarkReconciled := (quark.Spec.Version == quark.Status.Version)

	if quarkReconciled &&
		quark.Status.Health.Operator == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.OperatorReady = true
	} else {
		ntapQuarkStatus.OperatorReady = false
	}

	if quarkReconciled &&
		quark.Status.Health.NVC == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.VolumeControllerReady = true
	} else {
		ntapQuarkStatus.VolumeControllerReady = false
	}

	if quarkReconciled &&
		quark.Status.Health.NHC == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.HighAvailabilityReady = true
	} else {
		ntapQuarkStatus.HighAvailabilityReady = false
	}

	if quarkReconciled &&
		quark.Status.Health.NSC == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.ExternalVolumeAccessReady = true
	} else {
		ntapQuarkStatus.ExternalVolumeAccessReady = false
	}

	if quarkReconciled &&
		quark.Status.Health.CSC == quarkv1alpha1.QuarkReadyStatusReady && quark.Status.Health.ADC == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.CloudsnapshotCapabilityReady = true
	} else {
		ntapQuarkStatus.CloudsnapshotCapabilityReady = false
	}

	if quarkReconciled &&
		quark.Status.Health.IPD == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.IPDReady = true
	} else {
		ntapQuarkStatus.IPDReady = false
	}

	if quarkReconciled &&
		quark.Status.Health.EtcdCluster == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.EtcdClusterReady = true
	} else {
		ntapQuarkStatus.EtcdClusterReady = false
	}

	if quarkReconciled &&
		quark.Status.Health.Nodevol == quarkv1alpha1.QuarkReadyStatusReady {
		ntapQuarkStatus.NodevolReady = true
	} else {
		ntapQuarkStatus.NodevolReady = false
	}

	ntapQuarkStatus.Version = quark.Spec.Version

	return ntapQuarkStatus
}

func (client *NVCClient) NetAppVolumeCloudSnapshotGet(cloudSnapshotUUID string) (NetAppVolumeCloudSnapshotInfo, error) {
	cloudSnapshot, err := client.getCloudSnapshot(cloudSnapshotUUID)
	if err != nil {
		return NetAppVolumeCloudSnapshotInfo{}, err
	}
	return getCloudSnapshotInfo(cloudSnapshot), nil
}

func (client *NVCClient) NetAppVolumeCloudSnapshotGetByVolume(volumeUUID string) (cloudSnapshotList []NetAppVolumeCloudSnapshotInfo, err error) {
	allCloudSnapshotList, err := client.NetAppVolumeCloudSnapshotGetAll()
	if err != nil {
		return
	}

	for i := range allCloudSnapshotList {
		if allCloudSnapshotList[i].VolumeUUID == volumeUUID {
			cloudSnapshotList = append(cloudSnapshotList, allCloudSnapshotList[i])
		}
	}
	return
}

func (client *NVCClient) NetAppVolumeCloudSnapshotGetAll() ([]NetAppVolumeCloudSnapshotInfo, error) {
	listResult, err := client.quarkClient.QuarkV1alpha1().NetAppVolumeCloudSnapshots(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}

	var cloudSnapshotList []NetAppVolumeCloudSnapshotInfo

	for i := range listResult.Items {
		cloudSnapshotList = append(cloudSnapshotList, getCloudSnapshotInfo(&listResult.Items[i]))
	}

	return cloudSnapshotList, nil
}

func ConstructSecret(secretName string, data map[string][]byte) *corev1.Secret {
	return &corev1.Secret{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: secretName,
		},
		Data: data,
		Type: corev1.SecretTypeOpaque,
	}
}

func ConstructBucketCredentialsSecret(secretName, accessKey, secret string) *corev1.Secret {
	data := make(map[string][]byte)
	data["accessKey"] = []byte(accessKey)
	data["secret"] = []byte(secret)
	return ConstructSecret(secretName, data)
}

// Constructs SMB NetAppVolume definition.
// ownerID: OwnerID of the volume.
// dispalyName: Name of the volume provided by customer.
// sizeString: Volume Size.
// volumeUUID: Unique ID of volume in k8s cluster.
// volumePath: Volume mount path.
// serviceLevel: Standard for alpha.
// snapshotPolicy: Snapshot Policy object
// netbios: NetBIOS name
// domain: Active directory domain
// organizationUnit: Organization Unit
// dnsIPs: array of DNS IPs
// encryptionEnabled: Encryption enabled/disabled
func ConstructNetAppVolumeSMB(ownerID, displayName, sizeString, volumeUUID, volumePath, snapshotPolicyName string, serviceLevel ServiceLevel, netbios, domain, organizationUnit string, dnsIPs []string, encryptionEnabled bool) (volume *quarkv1alpha1.NetAppVolume, err error) {
	err = nvc.ValidateSMBConfiguration(volumePath, netbios, domain, organizationUnit, "")
	if err != nil {
		return nil, err
	}
	volume, err = constructBaseNetAppVolume(ownerID, displayName, sizeString, volumeUUID, volumePath, snapshotPolicyName, serviceLevel)
	if err != nil {
		return nil, err
	}

	volume.Spec.Protocol = "cifs"
	volume.Spec.SMBConfiguration.Netbios = netbios
	volume.Spec.SMBConfiguration.Domain = domain
	volume.Spec.SMBConfiguration.OrganizationUnit = organizationUnit
	volume.Spec.SMBConfiguration.EncryptionEnabled = encryptionEnabled
	volume.Spec.SMBConfiguration.DNSIPs = dnsIPs
	volume.Spec.SMBPasswordResetCount = 0

	return volume, nil
}

func (client *NVCClient) SecretCreate(secret *corev1.Secret) (*corev1.Secret, error) {
	return client.kubeClient.CoreV1().Secrets(client.nvcNamespace).Create(
		context.TODO(), secret, metav1.CreateOptions{})
}

func (client *NVCClient) SecretGet(secretName string) (*corev1.Secret, error) {
	return client.kubeClient.CoreV1().Secrets(client.nvcNamespace).Get(
		context.TODO(), secretName, metav1.GetOptions{})
}

func (client *NVCClient) SecretUpdate(secret *corev1.Secret) (*corev1.Secret, error) {
	return client.kubeClient.CoreV1().Secrets(client.nvcNamespace).Update(
		context.TODO(), secret, metav1.UpdateOptions{})
}

func (client *NVCClient) SecretCreateOrUpdate(secret *corev1.Secret) (*corev1.Secret, error) {
	newSecret, err := client.SecretCreate(secret)
	if err != nil && errors.IsAlreadyExists(err) {
		newSecret, err = client.SecretGet(secret.Name)
		if err == nil {
			newSecret.Data = secret.Data
			newSecret, err = client.SecretUpdate(newSecret)
		}
	}
	return newSecret, err
}

func (client *NVCClient) SecretDelete(secretName string) error {
	return client.kubeClient.CoreV1().Secrets(client.nvcNamespace).Delete(
		context.TODO(), secretName, metav1.DeleteOptions{})
}

func makeResolvConfString(dnsIPs []string) string {
	resolvData := ""
	for _, dnsIP := range dnsIPs {
		resolvData += "nameserver " + dnsIP + "\n"
	}
	return resolvData
}

func makeDNSConfigMap(dnsIPs []string) (*corev1.ConfigMap, error) {
	for i := range dnsIPs {
		dnsIPs[i] = strings.TrimSpace(dnsIPs[i])
		if net.ParseIP(dnsIPs[i]) == nil {
			return nil, fmt.Errorf("invalid dnsIP: %s", dnsIPs[i])
		}
	}
	sort.Strings(dnsIPs)
	resolvData := makeResolvConfString(dnsIPs)
	cfgMap := &corev1.ConfigMap{
		Data: map[string]string{
			"resolv.conf": resolvData,
		},
	}
	cfgMap.Name = quarkconfig.SMBDNSConfigName
	return cfgMap, nil
}

func (client *NVCClient) GetSMBDNSConfiguration() (*corev1.ConfigMap, error) {
	return client.kubeClient.CoreV1().ConfigMaps(client.nvcNamespace).Get(
		context.TODO(), quarkconfig.SMBDNSConfigName, metav1.GetOptions{})
}
func (client *NVCClient) CreateSMBDNSConfiguration(dnsIPs []string) error {
	if len(dnsIPs) == 0 {
		return fmt.Errorf("dnsIPs cannot be empty")
	}
	return nil
}

// This routine updates the existing DNS configmap and also updates
// DNS IP of existing NetAppVolumes
func (client *NVCClient) UpdateSMBDNSConfigurationAll(dnsIPs []string) error {
	if len(dnsIPs) == 0 {
		return fmt.Errorf("dnsIPs cannot be empty")
	}
	volumes, err := client.NetAppVolumeGetAll()
	if err != nil {
		return err
	}
	for _, volume := range volumes {
		if volume.Spec.Protocol == "cifs" {
			_, err = client.netAppVolumeSMBDNSUpdate(volume, dnsIPs)
			if err != nil {
				return err
			}
		}
	}
	return err
}

func (client *NVCClient) DeleteSMBDNSConfiguration() error {
	volumes, err := client.NetAppVolumeGetAllSMB()
	if err != nil {
		return err
	}
	if len(volumes) > 0 {
		return fmt.Errorf("failed to delete DNS configuration: still in use")
	}
	return nil
}

// GetKubeClient returns the current kubeclient for NVCClient
func (client *NVCClient) GetKubeClient() kubernetes.Interface {
	return client.kubeClient
}

// GetKubeClient returns the current kubeclient for NVCClient
func (client *NVCClient) GetQuarkClient() quarkclientset.Interface {
	return client.quarkClient
}

func (client *NVCClient) NetAppGetZoneStatus() (map[string]string, error) {
	var (
		err error
	)

	regionalHACM, err := client.kubeClient.CoreV1().ConfigMaps(
		client.nvcNamespace,
	).Get(context.TODO(), quarkconfig.RegionalHAConfigMapName, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("unable to get regional HA configmap")
	}

	return regionalHACM.Data, err
}

func (client *NVCClient) NetAppSetZoneStatusActive(zone string) error {
	var (
		err error
	)

	regionalHACM, err := client.kubeClient.CoreV1().ConfigMaps(
		client.nvcNamespace,
	).Get(context.TODO(), quarkconfig.RegionalHAConfigMapName, metav1.GetOptions{})
	if err != nil {
		return fmt.Errorf("unable to get regional HA configmap")
	}

	_, found := regionalHACM.Data[zone]
	if found == false {
		return fmt.Errorf("zone %s not found", zone)
	}

	// Mark the zone active
	cmCopy := regionalHACM.DeepCopy()
	cmCopy.Data[zone] = fmt.Sprintf("%s",
		quarkconfig.RegionalHAActiveZone)
	if regionalHACM, err = client.kubeClient.CoreV1().ConfigMaps(client.nvcNamespace).Update(
		context.TODO(), cmCopy, metav1.UpdateOptions{},
	); err != nil {
		log.WithFields(log.Fields{
			"error": err.Error(),
		}).Error("Failed to update regional HA configmap.")
		return fmt.Errorf("Failed to update zone status active")
	}

	log.WithFields(log.Fields{
		"zone": zone,
	}).Info("Updated zone status active")
	return nil
}
