package vbeta

import (
	"context"
	"encoding/json"
	"fmt"
	"strconv"
	"strings"

	log "github.com/sirupsen/logrus"

	appsv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"
	rbachelpers "k8s.io/api/rbac/v1"

	// API machinery
	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	types "k8s.io/apimachinery/pkg/types"

	// APIExtensions
	apiextensionsv1beta1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1beta1"

	// quark APIs
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

type patchStringValue struct {
	Op    string `json:"op"`
	Path  string `json:"path"`
	Value string `json:"value"`
}

func (client *NVCClient) ReconcileQuarkOperator(pullSecret string, registry string, version string) error {
	var err error
	if pullSecret != "" {
		err = client.checkSecretPresent(client.nvcNamespace, pullSecret)
		if err != nil {
			return fmt.Errorf("failed to get secret %s: %s", pullSecret, err.Error())
		}
	}

	err = client.reconcileQuarkCRD()
	if err != nil {
		return fmt.Errorf("failed to reconcile Quark CRD: %s", err.Error())
	}

	err = client.reconcileQuarkOperatorRBAC(pullSecret, registry, client.nvcNamespace)
	if err != nil {
		return fmt.Errorf("failed to reconcile RBAC for Quark Operator: %s", err.Error())
	}

	err = client.reconcileQuarkOperatorDeployment(registry, version)
	if err != nil {
		return fmt.Errorf("failed to reconcile quark operator deployment: %s", err.Error())
	}
	return nil
}

func (client *NVCClient) reconcileQuarkCRD() error {
	var err error
	qCrd := makeQuarkCRD()
	qCrdCur, err := client.crdClient.ApiextensionsV1beta1().CustomResourceDefinitions().Get(
		context.TODO(), qCrd.Name, metav1.GetOptions{})
	if err != nil {
		if errors.IsNotFound(err) {
			_, err := client.crdClient.ApiextensionsV1beta1().CustomResourceDefinitions().Create(
				context.TODO(), qCrd, metav1.CreateOptions{})
			if err != nil {
				return fmt.Errorf("failed to create Quark CRD: %s", err.Error())
			}
		} else {
			return fmt.Errorf("failed to get Quark CRD: %s", err.Error())
		}
	} else {
		qCrdCurrCopy := qCrdCur.DeepCopy()
		qCrd.ResourceVersion = qCrdCurrCopy.ResourceVersion
		_, err := client.crdClient.ApiextensionsV1beta1().CustomResourceDefinitions().Update(
			context.TODO(), qCrd, metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to update Quark CRD: %s", err.Error())
		}
	}
	return nil
}

func (client *NVCClient) checkSecretPresent(ns, secretname string) error {
	_, err := client.kubeClient.CoreV1().Secrets(ns).Get(
		context.TODO(), secretname, metav1.GetOptions{})
	if err != nil {
		return fmt.Errorf("failed to get secret %s: %s", secretname, err.Error())
	}
	return nil
}

func (client *NVCClient) reconcileServiceAccount(sa *corev1.ServiceAccount) error {
	saCurr, err := client.kubeClient.CoreV1().ServiceAccounts(sa.Namespace).Get(
		context.TODO(), sa.Name, metav1.GetOptions{})
	if err != nil {
		if errors.IsNotFound(err) {
			_, err := client.kubeClient.CoreV1().ServiceAccounts(sa.Namespace).Create(
				context.TODO(), sa, metav1.CreateOptions{})
			if err != nil {
				return fmt.Errorf("failed to create service account: %s, err: %s", sa.Name, err.Error())
			}
		} else {
			return fmt.Errorf("failed to get service account: %s, err: %s", sa.Name, err.Error())
		}
	} else {
		// We are Only supporting a Single Image-Pull-Secret
		doPatch := false
		var oldSecretName, newSecretName string
		var payload []patchStringValue

		if len(saCurr.ImagePullSecrets) > 0 {
			oldSecretName = saCurr.ImagePullSecrets[0].Name
		}
		if len(sa.ImagePullSecrets) > 0 {
			newSecretName = sa.ImagePullSecrets[0].Name
		}
		if oldSecretName != "" && newSecretName != "" && oldSecretName != newSecretName {
			doPatch = true
			// We are here because we originally had a different secret
			// now we are replacing that
			payload = []patchStringValue{{
				Op:    "replace",
				Path:  "/imagePullSecrets/0/name",
				Value: newSecretName,
			}}
		}

		if doPatch {
			payloadBytes, err := json.Marshal(payload)
			if err != nil {
				return fmt.Errorf("failed to marshal payload for patching image-pull-secret")
			}
			_, err = client.kubeClient.CoreV1().ServiceAccounts(sa.Namespace).Patch(
				context.TODO(), sa.Name, types.JSONPatchType, payloadBytes, metav1.PatchOptions{})
			if err != nil {
				return fmt.Errorf("failed to patch image-prepull-secret for service account: %s, err: %s", sa.Name, err.Error())
			}
		} else if oldSecretName != newSecretName {
			_, err = client.kubeClient.CoreV1().ServiceAccounts(sa.Namespace).Update(
				context.TODO(), sa, metav1.UpdateOptions{})
			if err != nil {
				return fmt.Errorf("failed to update image-prepull-secret for service account: %s, err: %s", sa.Name, err.Error())
			}
		}
	}
	return nil
}

func (client *NVCClient) reconcileQuarkOperatorRBAC(pullSecret, registry, namespace string) error {
	// setup service account
	sa := &corev1.ServiceAccount{
		ObjectMeta: metav1.ObjectMeta{
			Name:      "quark-operator",
			Namespace: namespace,
		},
	}

	// Setup ImagePullSecrets if required
	if pullSecret != "" {
		sa.ImagePullSecrets = []corev1.LocalObjectReference{{Name: pullSecret}}
	}

	err := client.reconcileServiceAccount(sa)
	if err != nil {
		return fmt.Errorf("failed to reconcile service account %s: %s", sa.Name, err.Error())
	}

	// Patch default service account in QuarkSystemNamespace with ImagePullSecrets
	sa = &corev1.ServiceAccount{
		ObjectMeta: metav1.ObjectMeta{
			Name:      "default",
			Namespace: quarkconfig.QuarkSystemNamespace,
		},
	}
	if pullSecret != "" {
		sa.ImagePullSecrets = []corev1.LocalObjectReference{{Name: pullSecret}}
	}

	err = client.reconcileServiceAccount(sa)
	if err != nil {
		return fmt.Errorf("failed to reconcile service account %s: %s", sa.Name, err.Error())
	}

	// setup quark-operator role
	cr := &rbachelpers.ClusterRole{
		ObjectMeta: metav1.ObjectMeta{
			Name: "quark-operator",
		},
		Rules: []rbachelpers.PolicyRule{
			{
				APIGroups: []string{""},
				Resources: []string{"serviceaccounts"},
				Verbs:     []string{"get", "list", "create", "delete", "watch", "patch", "update"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"pods", "nodes", "nodes/proxy", "nodes/metrics", "nodes/status"},
				Verbs:     []string{"get", "list", "watch", "update", "patch", "delete"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"ingresses"},
				Verbs:     []string{"get", "list", "watch"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"persistentvolumeclaims", "persistentvolumes"},
				Verbs:     []string{"get", "create", "list", "watch", "update", "delete"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"configmaps", "services", "endpoints", "namespaces", "events"},
				Verbs:     []string{"get", "list", "create", "watch", "update", "delete", "patch"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"limitranges", "replicationcontrollers", "resourcequotas"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"apps"},
				Resources: []string{"deployments", "daemonsets", "replicasets", "statefulsets"},
				Verbs:     []string{"get", "list", "watch", "create", "update", "delete"},
			},
			{
				APIGroups: []string{"apps"},
				Resources: []string{"deployments"},
				Verbs:     []string{"*"},
			},
			{
				APIGroups: []string{"etcd.database.coreos.com"},
				Resources: []string{"etcdclusters", "etcdbackups", "etcdrestores"},
				Verbs:     []string{"*"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"pods", "services", "endpoints", "secrets", "persistentvolumeclaims", "events"},
				Verbs:     []string{"*"},
			},
			{
				APIGroups: []string{"storage.k8s.io"},
				Resources: []string{"storageclasses", "volumeattachments"},
				Verbs:     []string{"get", "list", "watch", "create", "update", "delete"},
			},
			{
				APIGroups: []string{"apiextensions.k8s.io"},
				Resources: []string{"customresourcedefinitions"},
				Verbs:     []string{"*"},
			},
			{
				APIGroups: []string{"rbac.authorization.k8s.io"},
				Resources: []string{"roles", "rolebindings", "clusterroles", "clusterrolebindings"},
				Verbs:     []string{"get", "create", "delete", "update"},
			},
			{
				APIGroups: []string{"admissionregistration.k8s.io"},
				Resources: []string{"mutatingwebhookconfigurations", "validatingwebhookconfigurations"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"authentication.k8s.io"},
				Resources: []string{"tokenreviews"},
				Verbs:     []string{"create", "delete", "update"},
			},
			{
				APIGroups: []string{"authorization.k8s.io"},
				Resources: []string{"subjectaccessreviews"},
				Verbs:     []string{"create", "delete", "update"},
			},
			{
				APIGroups: []string{"autoscaling"},
				Resources: []string{"horizontalpodautoscalers"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"batch"},
				Resources: []string{"cronjobs", "jobs"},
				Verbs:     []string{"get", "list", "watch", "create", "update", "delete"},
			},
			{
				APIGroups: []string{"certificates.k8s.io"},
				Resources: []string{"certificatesigningrequests"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"extensions"},
				Resources: []string{"ingress", "ingresses", "ingresses/status"},
				Verbs:     []string{"get", "list", "watch"},
			},
			{
				APIGroups: []string{"networking.k8s.io"},
				Resources: []string{"networkpolicies"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"policy"},
				Resources: []string{"poddisruptionbudgets"},
				Verbs: []string{
					"list", "watch", "get", "create", "delete", "update",
				},
			},
			{
				APIGroups: []string{"scheduling.k8s.io"},
				Resources: []string{"priorityclasses"},
				Verbs: []string{
					"list", "watch", "get", "create", "delete", "update",
				},
			},
			{
				NonResourceURLs: []string{"/metrics"},
				Verbs:           []string{"get"},
			},
			{
				APIGroups: []string{"coordination.k8s.io"},
				Resources: []string{"leases"},
				Verbs:     []string{"get", "list", "watch"},
			},
			{
				APIGroups: []string{"quark.netapp.io"},
				Resources: []string{"*"},
				Verbs:     []string{"*"},
			},
		},
	}
	_, err = client.kubeClient.RbacV1().ClusterRoles().Get(
		context.TODO(), cr.Name, metav1.GetOptions{})
	if err != nil {
		if errors.IsNotFound(err) {
			_, err = client.kubeClient.RbacV1().ClusterRoles().Create(
				context.TODO(), cr, metav1.CreateOptions{})
			if err != nil {
				return fmt.Errorf("failed to create quark-operator cluster-role: %s", err.Error())
			}
		} else {
			return fmt.Errorf("failed to get quark-operator cluster-role: %s", err.Error())
		}
	} else {
		_, err = client.kubeClient.RbacV1().ClusterRoles().Update(
			context.TODO(), cr, metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to update quark-operator cluster-role: %s", err.Error())
		}
	}

	// setup quark-operator cluster-role binding
	crb := &rbachelpers.ClusterRoleBinding{
		ObjectMeta: metav1.ObjectMeta{
			Name: "quark-operator",
		},
		Subjects: []rbachelpers.Subject{
			{
				Kind:      "ServiceAccount",
				Name:      "quark-operator",
				Namespace: client.nvcNamespace,
			},
		},
		RoleRef: rbachelpers.RoleRef{
			APIGroup: "rbac.authorization.k8s.io",
			Kind:     "ClusterRole",
			Name:     "quark-operator",
		},
	}

	_, err = client.kubeClient.RbacV1().ClusterRoleBindings().Get(
		context.TODO(), crb.Name, metav1.GetOptions{})
	if err != nil {
		if errors.IsNotFound(err) {
			_, err = client.kubeClient.RbacV1().ClusterRoleBindings().Create(
				context.TODO(), crb, metav1.CreateOptions{})
			if err != nil {
				return fmt.Errorf("failed to create quark-operator cluster-role-binding: %s", err.Error())
			}
		} else {
			return fmt.Errorf("failed to get quark-operator cluster-role-binding: %s", err.Error())
		}
	} else {
		_, err = client.kubeClient.RbacV1().ClusterRoleBindings().Update(
			context.TODO(), crb, metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to update quark-operator cluster-role-binding: %s", err.Error())
		}
	}
	return nil
}

func (client *NVCClient) reconcileQuarkOperatorDeployment(registry string, version string) error {
	var err error
	dep := makeQuarkOperatorDeployment(client.nvcNamespace, registry, version)
	_, err = client.kubeClient.AppsV1().Deployments(dep.Namespace).Get(
		context.TODO(), dep.Name, metav1.GetOptions{})

	if err != nil {
		if errors.IsNotFound(err) {
			_, err = client.kubeClient.AppsV1().Deployments(dep.Namespace).Create(
				context.TODO(), dep, metav1.CreateOptions{})
			if err != nil {
				return fmt.Errorf("failed to create deployment: %s: %s", dep.Name, err.Error())
			}
		} else {
			return fmt.Errorf("failed to get deployment: %s: %s", dep.Name, err.Error())
		}
	} else {
		_, err = client.kubeClient.AppsV1().Deployments(dep.Namespace).Update(
			context.TODO(), dep, metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to update deployment: %s: %s", dep.Name, err.Error())
		}
	}
	return nil
}

func makeQuarkCRD() *apiextensionsv1beta1.CustomResourceDefinition {
	qCrd := &apiextensionsv1beta1.CustomResourceDefinition{
		ObjectMeta: metav1.ObjectMeta{
			Name: "quarks.quark.netapp.io",
		},
		Spec: apiextensionsv1beta1.CustomResourceDefinitionSpec{
			Group: "quark.netapp.io",
			Scope: apiextensionsv1beta1.NamespaceScoped,
			Subresources: &apiextensionsv1beta1.CustomResourceSubresources{
				Status: &apiextensionsv1beta1.CustomResourceSubresourceStatus{},
			},
			Versions: []apiextensionsv1beta1.CustomResourceDefinitionVersion{
				{
					Name:    "v1alpha1",
					Served:  true,
					Storage: true,
				},
			},
			Names: apiextensionsv1beta1.CustomResourceDefinitionNames{
				Kind:     "Quark",
				ListKind: "QuarkList",
				Plural:   "quarks",
				Singular: "quark",
			},
			AdditionalPrinterColumns: []apiextensionsv1beta1.CustomResourceColumnDefinition{
				{
					Name:        "Version",
					Type:        "string",
					Description: "Quark Spec Version",
					JSONPath:    ".spec.version",
				},
				{
					Name:        "Status",
					Type:        "string",
					Description: "Quark Status Version",
					JSONPath:    ".status.version",
				},
			},
		},
	}

	return qCrd
}

func makeQuarkOperatorDeployment(ns string, registry string, version string) *appsv1.Deployment {
	replicas := int32(1)
	quarkopImg := fmt.Sprintf("%s/quark-operator:%s", registry, version)

	dep := &appsv1.Deployment{
		ObjectMeta: metav1.ObjectMeta{
			Name:      quarkconfig.QuarkOperatorDeploymentName,
			Namespace: ns,
		},
		Spec: appsv1.DeploymentSpec{
			Replicas: &replicas,
			Selector: &metav1.LabelSelector{
				MatchLabels: map[string]string{
					"app": quarkconfig.QuarkOperatorDeploymentName,
				},
			},
			Template: corev1.PodTemplateSpec{
				ObjectMeta: metav1.ObjectMeta{
					Labels: map[string]string{
						"app": quarkconfig.QuarkOperatorDeploymentName,
					},
				},
				Spec: corev1.PodSpec{
					ServiceAccountName: quarkconfig.QuarkOperatorDeploymentName,
					Containers: []corev1.Container{
						{
							Image:           quarkopImg,
							ImagePullPolicy: corev1.PullIfNotPresent,
							Name:            quarkconfig.QuarkOperatorDeploymentName,
							Args: []string{
								"--namespace=$(NAMESPACE)",
							},
							Resources: quarkconfig.GetQuarkInfraContainerResources(quarkconfig.QuarkContainerQuarkOp, quarkconfig.CloudUnknown),
							Env: []corev1.EnvVar{
								{
									Name: "NAMESPACE",
									ValueFrom: &corev1.EnvVarSource{
										FieldRef: &corev1.ObjectFieldSelector{
											FieldPath: "metadata.namespace",
										},
									},
								},
							},
						},
					},
				},
			},
			Strategy: appsv1.DeploymentStrategy{
				Type: appsv1.RecreateDeploymentStrategyType,
			},
		},
	}
	return dep
}

// NetAppQuarkOperatorReady will return if Quark is in ready state
func (client *NVCClient) NetAppQuarkOperatorReady(quarkCR *quarkv1alpha1.Quark) bool {
	quarkStats := getNetAppQuarkStatus(quarkCR)
	// check for quark-operator
	return quarkStats.OperatorReady
}

// NetAppQuarkReady will return if Quark is in ready state
func (client *NVCClient) NetAppQuarkReady(quarkCR *quarkv1alpha1.Quark) bool {
	quarkStats := getNetAppQuarkStatus(quarkCR)

	// check for operator ready
	if !quarkStats.OperatorReady {
		return false
	}

	// check for netapp-volume-controller
	if !quarkStats.VolumeControllerReady {
		return false
	}

	// check for netapp-ha-controller
	if quarkCR.Spec.Cloud == quarkconfig.CloudGCP && !quarkStats.HighAvailabilityReady {
		return false
	}

	if !quarkStats.IPDReady {
		return false
	}

	if !quarkStats.EtcdClusterReady {
		return false
	}

	if strings.EqualFold(quarkCR.Spec.Nodevol, "true") {
		if !quarkStats.NodevolReady {
			return false
		}
	}

	// check for ntapvol-snat-controller
	if !quarkStats.ExternalVolumeAccessReady {
		return false
	}

	// checks for cloud-snapshot-controllers
	if !quarkStats.CloudsnapshotCapabilityReady {
		return false
	}

	if quarkCR.Spec.Version != quarkCR.Status.Version {
		return false
	}

	// all odds won, good to go
	return true
}

// NetAppQuarkGet will get Quark CR in the nvcClient Namespace in the cluster
func (client *NVCClient) NetAppQuarkGet(quarkCRName string) (*quarkv1alpha1.Quark, error) {
	return client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).Get(
		context.TODO(), quarkCRName, metav1.GetOptions{})
}

// NetAppQuarkGetAll will get all Quark CRs in the nvcClient Namespace in the cluster
func (client *NVCClient) NetAppQuarkGetAll() ([]*quarkv1alpha1.Quark, error) {
	quarks := make([]*quarkv1alpha1.Quark, 0)
	quarkList, err := client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, fmt.Errorf("failed to list quarks in cluster: %s", err.Error())
	}
	for i := range quarkList.Items {
		quarks = append(quarks, &quarkList.Items[i])
	}
	return quarks, nil
}

// NetAppQuarkCreate will attempt to create a Quark CR in the nvcClient Namespace in the cluster
func (client *NVCClient) NetAppQuarkCreate(quarkCR *quarkv1alpha1.Quark) (*quarkv1alpha1.Quark, error) {

	quarkList, err := client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).List(context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, fmt.Errorf("failed to list quark in cluster: %s", err.Error())
	}

	// Get List of Quarks in cluster and fail new creation if one exists
	if len(quarkList.Items) != 0 {
		if strings.EqualFold(quarkList.Items[0].Name, quarkCR.Name) {
			return nil, fmt.Errorf("failed to create quark, %s already exists", quarkCR.Name)
		}
		return nil, fmt.Errorf("failed to create quark, operation not allowed")
	}

	// Validations for QuarkCR
	switch quarkCR.Spec.Cloud {
	case quarkconfig.CloudGCP:
	case quarkconfig.CloudCTL:
	default:
		return nil, fmt.Errorf("failed to create quark, invalid cloud specified")
	}

	// Validations for Variant
	switch quarkCR.Spec.ControllerConfig.Variant {
	case quarkconfig.ProductionVariant:
	case quarkconfig.OnPremVariant:
	default:
		return nil, fmt.Errorf("failed to create quark, invalid variant")
	}

	return client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).Create(
		context.TODO(), quarkCR, metav1.CreateOptions{})
}

// NetAppQuarkUpdate will attempt to update the given Quark CR in the nvcClient Namespace in the cluster
// This API assumes the CR passed is already created
func (client *NVCClient) NetAppQuarkUpdate(quarkCR *quarkv1alpha1.Quark) (*quarkv1alpha1.Quark, error) {
	newQuark, err := client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).Update(
		context.TODO(), quarkCR, metav1.UpdateOptions{})
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}

	return newQuark, nil
}

func (client *NVCClient) NetAppQuarkUpdateVersion(
	quarkCRName, newVersion string, nodevol string,
) (*quarkv1alpha1.Quark, error) {
	quarkCR, err := client.NetAppQuarkGet(quarkCRName)
	if err != nil {
		return nil, fmt.Errorf("failed to get quark: %s", err.Error())
	}

	// DeepCopy
	quarkCRCopy := quarkCR.DeepCopy()

	// get the registry info
	registry := strings.Split(quarkCR.Spec.Images.DMAP, "/quark_dmap")[0]

	// makeImageInfo the new Version tag for all NetApp-Built Images
	quarkCRCopy.Spec.Images = makeImageInfo(registry, newVersion,
		quarkCRCopy.Spec.Images.PullSecret, quarkCRCopy.Spec.Images.PullPolicy)

	// Update service levels.
	if quarkCRCopy.Spec.ServiceLevels == nil {
		if quarkCRCopy.Spec.StorageClass.Data == "" || quarkCRCopy.Spec.StorageClass.Quark == "" || quarkCRCopy.Spec.StorageClass.Nvram == "" {
			log.WithFields(log.Fields{
				"data":  quarkCRCopy.Spec.StorageClass.Data,
				"quark": quarkCRCopy.Spec.StorageClass.Quark,
				"nvram": quarkCRCopy.Spec.StorageClass.Nvram,
			}).Warn("Service levels are unspecified and not all PD types in " +
				"storageClass are available.")
		}
		// Note that ServiceLevels will only be nil for zonal clusters; regional
		// clusters will have them from the get-go.
		quarkCRCopy.Spec.ServiceLevels = []quarkv1alpha1.ServiceLevel{
			quarkv1alpha1.ServiceLevel{
				Name:     string(quarkconfig.ServiceLevelStandard),
				Regional: false,
				DataSC:   quarkCRCopy.Spec.StorageClass.Data,
				QuarkSC:  quarkCRCopy.Spec.StorageClass.Quark,
				NvramSC:  quarkCRCopy.Spec.StorageClass.Nvram,
			},
		}
	}
	//set etcd storageclass when upgrading
	quarkCRCopy.Spec.StorageClass.Etcd = quarkCRCopy.Spec.StorageClass.Nvram
	if quarkCRCopy.Spec.StorageClass.Etcd == "" {
		for _, sl := range quarkCRCopy.Spec.ServiceLevels {
			if sl.Name == string(quarkconfig.ServiceLevelStandard) {
				quarkCRCopy.Spec.StorageClass.Etcd = sl.NvramSC
			}
		}
	}
	//set nodevol storageclass when upgrading
	quarkCRCopy.Spec.StorageClass.Nodevol = quarkCRCopy.Spec.StorageClass.Nvram
	if quarkCRCopy.Spec.StorageClass.Nodevol == "" {
		for _, sl := range quarkCRCopy.Spec.ServiceLevels {
			if sl.Name == string(quarkconfig.ServiceLevelStandard) {
				quarkCRCopy.Spec.StorageClass.Nodevol = sl.NvramSC
			}
		}
	}

	// handling clusterwide etcd introduction
	contConfig := &quarkCRCopy.Spec.ControllerConfig
	contConfig.EtcdClusterSize = 3
	if contConfig.Variant == quarkconfig.OnPremVariant || (contConfig.Variant == quarkconfig.InternalVariant && quarkCRCopy.Spec.Cloud != quarkconfig.CloudGCP) {
		contConfig.EtcdClusterSize = 1
	}
	contConfig.EtcdVersion = "3.3.18-custom-v1"
	contConfig.EtcdRepository = fmt.Sprintf("%s/quark_etcd", registry)
	quarkCRCopy.Spec.Nodevol = nodevol

	// update the version
	quarkCRCopy.Spec.Version = newVersion

	newQuark, err := client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).Update(
		context.TODO(), quarkCRCopy, metav1.UpdateOptions{})
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}

	return newQuark, nil
}

// NetAppQuarkDelete will attempt to delete the given Quark CR in the nvcClient Namespace in the cluster
func (client *NVCClient) NetAppQuarkDelete(quarkCRName string) error {
	vols, err := client.NetAppVolumeGetAll()
	if err != nil {
		return fmt.Errorf("failed to get netappvolumes: %s", err.Error())
	}

	if len(vols) != 0 {
		return fmt.Errorf("%v netappvolumes exist in cluster, cannot delete quark", len(vols))
	}
	return client.quarkClient.QuarkV1alpha1().Quarks(client.nvcNamespace).Delete(
		context.TODO(), quarkCRName, metav1.DeleteOptions{})
}

func makeImageInfo(registry, version, pullSecretName, pullPolicy string) quarkv1alpha1.Images {
	imgInfo := quarkv1alpha1.Images{}

	imgInfo.PullPolicy = pullPolicy
	imgInfo.PullSecret = pullSecretName

	imgInfo.CSC = fmt.Sprintf("%s/cloud-snapshot-controller:%s", registry, version)
	imgInfo.NHC = fmt.Sprintf("%s/netapp-ha-controller:%s", registry, version)
	imgInfo.NVC = fmt.Sprintf("%s/netapp-volume-controller:%s", registry, version)
	imgInfo.NSC = fmt.Sprintf("%s/ntapvol-snat-controller:%s", registry, version)
	imgInfo.Podrick = fmt.Sprintf("%s/podrick:%s", registry, version)

	// Built in ONTAP
	imgInfo.ADC = fmt.Sprintf("%s/adc:%s", registry, version)
	imgInfo.DMAP = fmt.Sprintf("%s/quark_dmap:%s", registry, version)
	imgInfo.Secd = fmt.Sprintf("%s/quark_secd:%s", registry, version)
	imgInfo.Supportability = fmt.Sprintf("%s/quark_supportability:%s", registry, version)
	imgInfo.NodevolWorker = fmt.Sprintf("%s/nodevol-worker:%s", registry, version)
	imgInfo.NodevolController = fmt.Sprintf("%s/nodevol-controller:%s", registry, version)

	// External components in gcr
	imgInfo.Etcd = fmt.Sprintf("%s/quark_etcd:v3.3.18-custom-v1", registry)
	imgInfo.EventRouter = fmt.Sprintf("%s/eventrouter:v0.3", registry)
	imgInfo.Logging = fmt.Sprintf("%s/fluent-bit:1.2.2", registry)
	imgInfo.Busybox = fmt.Sprintf("%s/busybox:1.31.1", registry)
	imgInfo.InitContainer = fmt.Sprintf("%s/busybox:1.31.1", registry)

	// Cluster-wide Etcd
	imgInfo.EtcdInitContainer = fmt.Sprintf("%s/busybox:1.28.0-glibc", registry)
	imgInfo.EtcdOperator = fmt.Sprintf("%s/etcd-operator:v0.9.4-atom120Q2CY21", registry)

	// Kube-State-Metrics
	imgInfo.KubeMetrics = fmt.Sprintf("%s/kube-state-metrics:v1.9.5", registry)
	// PrometheusInfra
	imgInfo.Prometheus = fmt.Sprintf("%s/prometheus:v2.16.0", registry)
	imgInfo.PrometheusConfigMapReload = fmt.Sprintf("%s/configmap-reload:v0.2.2", registry)
	imgInfo.PrometheusExporter = fmt.Sprintf("%s/node-exporter:v0.18.0", registry)
	imgInfo.HASpare = fmt.Sprintf("%s/google_containers_pause:3.1", registry)
	// Startup script
	imgInfo.Startup = fmt.Sprintf("%s/google_containers_startup_script:v1", registry)

	return imgInfo
}

func (client *NVCClient) ConstructQuarkResource(name, version, cloud, registry, pullSecretName, pullPolicy string,
	secrets []quarkv1alpha1.Secret,
	storageClass *quarkv1alpha1.StorageClass,
	volume *quarkv1alpha1.Volume,
	ilb *quarkv1alpha1.ILB,
	images *quarkv1alpha1.Images,
	customerProject *quarkv1alpha1.Project,
	controllerConfig *quarkv1alpha1.ControllerConfig,
	logging *quarkv1alpha1.Logging,
	nodevol string,
) (*quarkv1alpha1.Quark, error) {
	return client.constructQuarkResourceFull(name, version, cloud, registry,
		pullSecretName, pullPolicy,
		secrets,
		storageClass,
		volume,
		ilb,
		images,
		customerProject,
		controllerConfig,
		logging,
		false,
		nodevol,
	)
}

func (client *NVCClient) ConstructRegionalQuarkResource(name, version, cloud, registry, pullSecretName, pullPolicy string,
	secrets []quarkv1alpha1.Secret,
	storageClass *quarkv1alpha1.StorageClass,
	volume *quarkv1alpha1.Volume,
	ilb *quarkv1alpha1.ILB,
	images *quarkv1alpha1.Images,
	customerProject *quarkv1alpha1.Project,
	controllerConfig *quarkv1alpha1.ControllerConfig,
	logging *quarkv1alpha1.Logging,
	nodevol string,
) (*quarkv1alpha1.Quark, error) {
	return client.constructQuarkResourceFull(name, version, cloud, registry,
		pullSecretName, pullPolicy,
		secrets,
		storageClass,
		volume,
		ilb,
		images,
		customerProject,
		controllerConfig,
		logging,
		true,
		nodevol,
	)
}

// ConstructQuarkResource will construct a Quark CR for caller to use in a CreateQuarkResource call
// TODO: Validations on the inputs
func (client *NVCClient) constructQuarkResourceFull(name, version, cloud, registry, pullSecretName, pullPolicy string,
	secrets []quarkv1alpha1.Secret,
	storageClass *quarkv1alpha1.StorageClass,
	volume *quarkv1alpha1.Volume,
	ilb *quarkv1alpha1.ILB,
	images *quarkv1alpha1.Images,
	customerProject *quarkv1alpha1.Project,
	controllerConfig *quarkv1alpha1.ControllerConfig,
	logging *quarkv1alpha1.Logging,
	regional bool,
	nodevol string,
) (*quarkv1alpha1.Quark, error) {

	// validate inputs
	if name == "" || version == "" || cloud == "" || registry == "" {
		return nil, fmt.Errorf("name, version, cloud & registry info required")
	}

	// use ILB info given and error if not
	ilbinfo := quarkv1alpha1.ILB{}
	if ilb == nil {
		return nil, fmt.Errorf("ilb info required")
	} else {
		ilb.DeepCopyInto(&ilbinfo)
	}

	logInfo := quarkv1alpha1.Logging{}
	if logging == nil {
		return nil, fmt.Errorf("logging info required")
	} else {
		logging.DeepCopyInto(&logInfo)
	}

	projInfo := quarkv1alpha1.Project{}
	if customerProject == nil {
		return nil, fmt.Errorf("CustomerProject info required")
	} else {
		customerProject.DeepCopyInto(&projInfo)
	}

	// use given or apply a default one
	contConfig := quarkv1alpha1.ControllerConfig{}
	if controllerConfig == nil {
		contConfig = quarkv1alpha1.ControllerConfig{
			AdcGetInterval:  "10",
			EtcdClusterSize: 3,
			EtcdVersion:     "3.3.18-custom-v1",
			EtcdRepository:  fmt.Sprintf("%s/quark_etcd", registry),
		}
	} else {
		controllerConfig.DeepCopyInto(&contConfig)
	}

	if contConfig.Variant == "" {
		contConfig.Variant = quarkconfig.ProductionVariant
	}

	imgInfo := quarkv1alpha1.Images{}
	if images == nil {
		imgInfo = makeImageInfo(registry, version, pullSecretName,
			pullPolicy)
	} else {
		images.DeepCopyInto(&imgInfo)
	}

	// use given or create new storage class
	// TODO: The passed in storage class settings for Nvram, Data, and Quark
	// only apply to the standard storage class, even in regional clusters;
	// they also show up in the Quark CR, even though they're not directly
	// used in setting up the cluster. We should deprecate these and set these
	// storage class values based on the variant.
	scInfo := quarkv1alpha1.StorageClass{}
	if storageClass == nil {
		scInfo = quarkv1alpha1.StorageClass{
			Provisioners: []quarkv1alpha1.SCProvisioner{
				{
					Provisioner:  "kubernetes.io/gce-pd",
					StorageClass: "pd-ssd",
					Regional:     false,
				},
			},
			Nvram: "pd-ssd",
			Data:  "pd-ssd",
			Quark: "pd-ssd",
			Etcd:  "pd-ssd",
		}
		if strings.EqualFold(nodevol, "true") {
			scInfo.Nodevol = "pd-ssd"
		}

		if regional {
			scInfo.Provisioners = append(scInfo.Provisioners,
				quarkv1alpha1.SCProvisioner{
					Provisioner:  "kubernetes.io/gce-pd",
					StorageClass: "pd-ssd",
					Name:         "regional-pd-ssd",
					Regional:     true,
				},
			)
			scInfo.Provisioners = append(scInfo.Provisioners,
				quarkv1alpha1.SCProvisioner{
					Provisioner:  "kubernetes.io/gce-pd",
					StorageClass: "pd-standard",
					Name:         "regional-pd-standard",
					Regional:     true,
				},
			)
		}
	} else {
		storageClass.DeepCopyInto(&scInfo)
	}

	serviceLevels := []quarkv1alpha1.ServiceLevel{
		quarkv1alpha1.ServiceLevel{
			Name:      string(quarkconfig.ServiceLevelStandard),
			NvramSC:   scInfo.Nvram,
			DataSC:    scInfo.Data,
			QuarkSC:   scInfo.Quark,
			Regional:  false,
			RightSize: true,
		},
	}
	if regional {
		serviceLevels = append(serviceLevels,
			quarkv1alpha1.ServiceLevel{
				Name:      string(quarkconfig.ServiceLevelRegionalStandard),
				NvramSC:   "regional-pd-ssd",
				DataSC:    "regional-pd-ssd",
				QuarkSC:   "regional-pd-ssd",
				Regional:  true,
				RightSize: true,
			},
		)
	}

	// use given or create new volume
	volInfo := quarkv1alpha1.Volume{}
	if volume == nil {
		volInfo = quarkv1alpha1.Volume{
			QuarkVolumeSize: quarkconfig.QuarkVolumeSize,
			NvramSize:       quarkconfig.NvramSize,
			MinPVCSize:      quarkconfig.MinPVCSize,
			MaxPVCSize:      quarkconfig.MaxPVCSize,
		}
	} else {
		volume.DeepCopyInto(&volInfo)
	}

	// populate secret info if specified
	secInfo := []quarkv1alpha1.Secret{}
	secInfo = append(secInfo, secrets...)

	quarkCR := &quarkv1alpha1.Quark{
		metav1.TypeMeta{},
		metav1.ObjectMeta{
			Name: name,
			Labels: map[string]string{
				"app": quarkconfig.QuarkOperatorDeploymentName,
			},
		},
		quarkv1alpha1.QuarkSpec{
			Version:          version,
			Namespace:        client.nvcNamespace,
			Cloud:            cloud,
			DarkContent:      strconv.FormatBool(client.DarkContent),
			Regional:         regional,
			Secrets:          secInfo,
			StorageClass:     scInfo,
			ServiceLevels:    serviceLevels,
			Volume:           volInfo,
			ILB:              ilbinfo,
			Images:           imgInfo,
			Project:          projInfo,
			ControllerConfig: contConfig,
			Logging:          logInfo,
			Nodevol:          nodevol,
		},
		quarkv1alpha1.QuarkStatus{},
	}
	return quarkCR, nil
}
